# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from itertools import tee
from typing import List

import numpy as np

from src.gluonts.dataset.common import Dataset
from src.gluonts.evaluation import Evaluator
from src.gluonts.model.predictor import Predictor

from src.gluonts.nursery.temporal_hierarchical_forecasting.utils.common import (
    TEMPORAL_HIERARCHIES,
)
from src.gluonts.nursery.temporal_hierarchical_forecasting.utils.utils import (
    get_ts_at_all_levels,
    unpack_forecasts,
)
from src.gluonts.nursery.temporal_hierarchical_forecasting.utils.gluonts_helper import (
    make_predictions,
    to_dataframe_it,
    truncate_target,
)


def evaluate_predictor(
    predictor: Predictor,
    test_dataset: Dataset,
    freq: str,
    metrics: List[str] = ["mean_wQuantileLoss"],
    evaluate_all_levels: bool = False,
):
    pred_input_it = truncate_target(
        test_dataset,
        prediction_length=predictor.prediction_length,
    )

    forecast_it = make_predictions(
        pred_input=pred_input_it,
        predictor=predictor,
        forecasts_at_all_levels=evaluate_all_levels,
    )
    test_ts_it = to_dataframe_it(test_dataset)

    evaluator = Evaluator(quantiles=(np.arange(20) / 20.0)[1:], num_workers=0)
    if not evaluate_all_levels:
        # Evaluate only the bottom level.
        agg_metrics, _ = evaluator(
            ts_iterator=test_ts_it,
            fcst_iterator=forecast_it,
        )
        metrics_to_return = {
            metric_str: agg_metrics[metric_str] for metric_str in metrics
        }
        return metrics_to_return

    # Evaluate all levels.
    metrics_to_return = evaluate_forecasts_at_all_levels(
        forecast_at_all_levels_it=forecast_it,
        test_ts_at_all_levels_it=get_ts_at_all_levels(
            ts_it=test_ts_it,
            temporal_hierarchy=TEMPORAL_HIERARCHIES[freq],
            prediction_length=predictor.prediction_length,
            target_temporal_hierarchy=TEMPORAL_HIERARCHIES[freq],
        ),
        temporal_hierarchy=TEMPORAL_HIERARCHIES[freq],
        evaluator=evaluator,
        metrics=metrics,
    )
    return metrics_to_return


def evaluate_forecasts_at_all_levels(
    forecast_at_all_levels_it,
    test_ts_at_all_levels_it,
    temporal_hierarchy,
    evaluator,
    metrics: List[str] = ["mean_wQuantileLoss"],
):
    # Order of forecasts obtained e.g., in case of the hierarchy: [6M, 2M, 1M]
    # (ts_1_6M, ts_1_2M, ts_1_1M, ts_2_6M, ts_2_2M, ts_2_1M, ...)
    forecast_at_all_levels_unpacked_it = unpack_forecasts(
        forecast_at_all_levels_it=forecast_at_all_levels_it,
        temporal_hierarchy=temporal_hierarchy,
        target_temporal_hierarchy=temporal_hierarchy,
    )

    num_levels = len(temporal_hierarchy.agg_multiples)

    # In one go, we can get item metrics for time series that have the same frequency.
    # So we create `num_levels` copies of the iterator and obtain the item metrics
    # for each level independently.
    forecast_at_all_levels_unpacked_it_set = tee(
        forecast_at_all_levels_unpacked_it,
        num_levels,
    )
    test_ts_at_all_levels_it_set = tee(test_ts_at_all_levels_it, num_levels)

    # Since forecasts for all granularities are in the same iterable,
    # we need a way to iterate through forecasts skipping some elements.
    def skip_iter(it, num_skips: int, offset: int):
        for _ in range(offset):
            next(it)
        for item in it:
            for _ in range(num_skips):
                next(it, None)  # None: in case the `it` is already exhausted.
            yield item

    metrics_to_return = {}
    for level in range(num_levels):
        agg_metrics_level, metrics_per_ts_level = evaluator(
            ts_iterator=skip_iter(
                test_ts_at_all_levels_it_set[level],
                num_skips=num_levels - 1,
                offset=level,
            ),
            fcst_iterator=skip_iter(
                forecast_at_all_levels_unpacked_it_set[level],
                num_skips=num_levels - 1,
                offset=level,
            ),
        )

        for metric_name in metrics:
            metrics_to_return[
                f"Freq_{temporal_hierarchy.freq_strs[level]}_{metric_name}"
            ] = agg_metrics_level[metric_name]

    return metrics_to_return



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple
import numpy as np

import torch
import torch.nn as nn
from src.gluonts.core.component import validated


class QuantileLoss(nn.Module):
    @validated()
    def __init__(
        self,
        quantiles: List[float],
        quantile_weights: Optional[List[float]] = None,
    ) -> None:
        super().__init__()

        self.quantiles = quantiles
        self.num_quantiles = len(quantiles)
        self.quantile_weights = (
            [1.0 / self.num_quantiles for i in range(self.num_quantiles)]
            if not quantile_weights
            else quantile_weights
        )

    def forward(
        self, y_true: torch.Tensor, y_pred: torch.Tensor, sample_weight=None
    ):
        if self.num_quantiles > 1:
            y_pred_all = torch.chunk(y_pred, self.num_quantiles, dim=-1)
        else:
            y_pred_all = [y_pred]

        qt_loss = []
        for i, y_pred_q in enumerate(y_pred_all):
            q = self.quantiles[i]
            weighted_qt = (
                self.compute_quantile_loss(y_true, y_pred_q.squeeze(-1), q)
                * self.quantile_weights[i]
            )
            qt_loss.append(weighted_qt)
        stacked_qt_losses = torch.stack(qt_loss, dim=-1)
        sum_qt_loss = torch.mean(stacked_qt_losses, dim=-1)
        if sample_weight is not None:
            return sample_weight * sum
        else:
            return sum_qt_loss

    @staticmethod
    def compute_quantile_loss(
        y_true: torch.Tensor, y_pred_p: torch.Tensor, p: float
    ) -> torch.Tensor:
        under_bias = p * torch.clamp(y_true - y_pred_p, min=0)
        over_bias = (1 - p) * torch.clamp(y_pred_p - y_true, min=0)

        qt_loss = 2 * (under_bias + over_bias)
        return qt_loss


class ProjectParams(nn.Module):
    @validated()
    def __init__(self, in_features, num_quantiles):
        super().__init__()
        self.projection = nn.Linear(
            in_features=in_features, out_features=num_quantiles
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.projection(x)


class QuantileOutput:
    @validated()
    def __init__(
        self,
        input_size,
        quantiles: List[float],
        quantile_weights: Optional[List[float]] = None,
    ) -> None:
        self.input_size = input_size
        self.quantiles = quantiles
        self.quantile_weights = quantile_weights

    def get_loss(self) -> nn.Module:
        return QuantileLoss(
            quantiles=self.quantiles, quantile_weights=self.quantile_weights
        )

    def get_quantile_proj(self) -> nn.Module:
        return ProjectParams(
            in_features=self.input_size, num_quantiles=len(self.quantiles)
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
from dataclasses import dataclass
from pathlib import Path
from typing import cast, List, Optional
from src.gluonts.mx import DeepAREstimator
from src.gluonts.model.estimator import DummyEstimator, Estimator
from src.gluonts.mx import NBEATSEstimator
from src.gluonts.model.naive_2 import Naive2Predictor
from src.gluonts.model.npts import NPTSEstimator
from src.gluonts.model.predictor import Predictor
from src.gluonts.model.prophet import ProphetPredictor
from src.gluonts.model.r_forecast import RForecastPredictor
from src.gluonts.model.seasonal_naive import SeasonalNaivePredictor
from src.gluonts.mx import MQCNNEstimator, MQRNNEstimator
from src.gluonts.mx import SimpleFeedForwardEstimator
from src.gluonts.mx.model.tft import TemporalFusionTransformerEstimator
from src.gluonts.mx.trainer.callback import Callback
from src.gluonts.time_feature import Constant
from mxnet.gluon import nn
from tsbench.config.dataset import DatasetConfig
from tsbench.config.dataset.datasets import WindFarmsDatasetConfig
from ._base import ModelConfig, TrainConfig
from ._factory import register_model


@register_model
@dataclass(frozen=True)
class SeasonalNaiveModelConfig(ModelConfig):
    """
    The seasonal naive estimator.
    """

    @classmethod
    def name(cls) -> str:
        return "seasonal_naive"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            SeasonalNaivePredictor,
            freq=freq,
            prediction_length=prediction_length,
        )


@register_model
@dataclass(frozen=True)
class ThetaModelConfig(ModelConfig):
    """
    The Theta R estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "theta"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            RForecastPredictor,
            freq=freq,
            prediction_length=prediction_length,
            method_name="thetaf",
        )


@dataclass(frozen=True)
class Naive2ModelConfig(ModelConfig):
    """
    The Naïve 2 model as used in the M4 competition.
    """

    @classmethod
    def name(cls) -> str:
        return "naive2"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            Naive2Predictor, freq=freq, prediction_length=prediction_length
        )


@register_model
@dataclass(frozen=True)
class NptsModelConfig(ModelConfig):
    """
    The NPTS estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "npts"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return NPTSEstimator(freq=freq, prediction_length=prediction_length)


@register_model
@dataclass(frozen=True)
class DeepARModelConfig(ModelConfig, TrainConfig):
    """
    The DeepAR estimator config.
    """

    num_layers: int = 2
    num_cells: int = 40

    @classmethod
    def name(cls) -> str:
        return "deepar"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        deepar_estimator = cast(DeepAREstimator, estimator)
        transform = deepar_estimator.create_transformation()
        return deepar_estimator.create_predictor(transform, network)

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DeepAREstimator(
            freq=freq,
            prediction_length=prediction_length,
            num_layers=self.num_layers,
            num_cells=self.num_cells,
            time_features=None if time_features else [],
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class MQCnnModelConfig(ModelConfig, TrainConfig):
    """
    The MQCNN estimator config.
    """

    context_length_multiple: int = 4
    num_filters: int = 30
    kernel_size_first: int = 7
    kernel_size_hidden: int = 3
    kernel_size_last: int = 3

    @classmethod
    def name(cls) -> str:
        return "mqcnn"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        mqcnn_estimator = cast(MQCNNEstimator, estimator)
        transform = mqcnn_estimator.create_transformation()
        return mqcnn_estimator.create_predictor(transform, cast)  # type: ignore

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return MQCNNEstimator(
            freq=freq,
            prediction_length=prediction_length,
            channels_seq=[self.num_filters] * 3,
            kernel_size_seq=[
                self.kernel_size_first,
                self.kernel_size_hidden,
                self.kernel_size_last,
            ],
            add_time_feature=time_features,
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class MQRnnModelConfig(ModelConfig, TrainConfig):
    """
    The MQRNN estimator config.
    """

    context_length_multiple: int = 4

    @classmethod
    def name(cls) -> str:
        return "mqrnn"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        mqrnn_estimator = cast(MQRNNEstimator, estimator)
        transform = mqrnn_estimator.create_transformation()
        return mqrnn_estimator.create_predictor(transform, network)  # type: ignore

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return MQRNNEstimator(
            freq=freq,
            prediction_length=prediction_length,
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class SimpleFeedforwardModelConfig(ModelConfig, TrainConfig):
    """
    The simple feedforward estimator config.
    """

    hidden_dim: int = 40
    num_layers: int = 2

    @classmethod
    def name(cls) -> str:
        return "simple_feedforward"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        ff_estimator = cast(SimpleFeedForwardEstimator, estimator)
        transform = ff_estimator.create_transformation()
        return ff_estimator.create_predictor(transform, network)

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return SimpleFeedForwardEstimator(
            freq=freq,
            prediction_length=prediction_length,
            num_hidden_dimensions=[self.hidden_dim] * self.num_layers,
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class TemporalFusionTransformerModelConfig(ModelConfig, TrainConfig):
    """
    The temporal fusion transformer estimator config.
    """

    hidden_dim: int = 32
    num_heads: int = 4

    @classmethod
    def name(cls) -> str:
        return "tft"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        tft_estimator = cast(TemporalFusionTransformerEstimator, estimator)
        transform = tft_estimator.create_transformation()
        return tft_estimator.create_predictor(transform, network)

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return TemporalFusionTransformerEstimator(
            freq=freq,
            prediction_length=prediction_length,
            hidden_dim=self.hidden_dim,
            num_heads=self.num_heads,
            num_outputs=9,
            time_features=[Constant()] if not time_features else [],
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
        )


@register_model
@dataclass(frozen=True)
class NBeatsModelConfig(ModelConfig, TrainConfig):
    """
    The NBeats estimator config.
    """

    context_length_multiple: int = 2
    num_stacks: int = 30
    num_blocks: int = 1

    @classmethod
    def name(cls) -> str:
        return "nbeats"

    @property
    def prediction_samples(self) -> int:
        return 1

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        nb_estimator = cast(NBEATSEstimator, estimator)
        transform = nb_estimator.create_transformation()
        return nb_estimator.create_predictor(transform, network)

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return NBEATSEstimator(
            freq=freq,
            prediction_length=prediction_length,
            num_stacks=self.num_stacks,
            num_blocks=[self.num_blocks],
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class ProphetModelConfig(ModelConfig):
    """
    The Prophet estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "prophet"

    def save_predictor(self, predictor: Predictor, path: Path) -> None:
        file = path / "metadata.pickle"
        with file.open("w") as f:
            json.dump(
                {
                    "freq": predictor.freq,
                    "prediction_length": predictor.prediction_length,
                },
                f,
            )

    def load_predictor(self, path: Path) -> Predictor:
        file = path / "metadata.pickle"
        with file.open("r") as f:
            meta = json.load(f)
        return ProphetPredictor(
            freq=meta["freq"], prediction_length=meta["prediction_length"]
        )

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            ProphetPredictor, freq=freq, prediction_length=prediction_length
        )


@register_model
@dataclass(frozen=True)
class STLARModelConfig(ModelConfig):
    """
    The STL-AR estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "stlar"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            RForecastPredictor,
            freq=freq,
            prediction_length=prediction_length,
            method_name="stlar",
        )


@register_model
@dataclass(frozen=True)
class ARIMAModelConfig(ModelConfig):
    """
    The ARIMA estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "arima"

    @property
    def prefers_parallel_predictions(self) -> bool:
        return True

    def max_time_series_length(self, config: DatasetConfig) -> Optional[int]:
        if isinstance(config, WindFarmsDatasetConfig):
            return 2880
        return 100000

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            RForecastPredictor,
            freq=freq,
            prediction_length=prediction_length,
            method_name="arima",
        )


@register_model
@dataclass(frozen=True)
class ETSModelConfig(ModelConfig):
    """
    The ETS estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "ets"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            RForecastPredictor,
            freq=freq,
            prediction_length=prediction_length,
            method_name="ets",
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional, List, Tuple, Union, Iterator, Callable
from itertools import product, chain
from copy import deepcopy
import warnings
import math

import torch as pt
from torch import Tensor, LongTensor, BoolTensor
from torch import nn
from torch.nn import Parameter, functional as F

from tslib.nn.activations import PositiveSoftplus
from ..network import (
    AttentionKernel,
    ExpKernel,
    RBFKernel,
    EncoderModule,
    DecoderModule,
    AttentionBlock,
    AdversarialBlock,
    SimpleDiscriminator,
)
from .hooks import (
    ZScoreNormalizer,
    LossFunction,
    MSELoss,
    MAELoss,
)

MIN_TRADEOFF = 0.5


class AttentionEstimator(nn.Module):
    def __init__(
        self,
        *lists_of_modules: List[nn.Module],
        horizon: Union[int, float],
        loss_func: LossFunction = MAELoss(),
        normalize_input: bool = True,
        rescale_output: bool = True,
        layerwise_loss: bool = True,
    ) -> None:
        super(AttentionEstimator, self).__init__()
        self._create_modules(*lists_of_modules)

        if isinstance(horizon, int):
            assert horizon >= 0, "Positive horizon is required."
        else:
            assert 0 <= horizon <= 1, "horizon must be from [0,1]."
        self._softplus = PositiveSoftplus(margin=MIN_TRADEOFF)
        self._tradeoff = Parameter(
            pt.tensor(math.log(math.exp(1 - MIN_TRADEOFF) - 1) + MIN_TRADEOFF),
            requires_grad=False,
        )
        self.horizon = horizon
        self.normalize_input = normalize_input
        self.rescale_output = rescale_output
        self.layerwise_loss = layerwise_loss

        if self.normalize_input:
            self._normalizer = ZScoreNormalizer(
                rescale_loss=self.rescale_output
            )
            self.register_forward_pre_hook(self._normalizer.forward_pre_hook)
            self.register_forward_hook(self._normalizer.forward_hook)
        else:
            self._normalizer = None
        self.register_loss_func(loss_func)

        self.register_buffer("bc_loss", None, persistent=False)
        self.register_buffer("fc_loss", None, persistent=False)
        self.register_buffer("residue", None, persistent=False)
        self.register_buffer("forecast", None, persistent=False)
        self.register_buffer("denominator", None, persistent=False)

    def _create_modules(self, *modules):
        self.blocks = nn.ModuleList(
            [AttentionBlock(*ms) for ms in zip(*modules)]
        )

    @property
    def tradeoff(self) -> Tensor:
        return self._softplus(self._tradeoff)

    @property
    def n_layer(self) -> int:
        return len(self.blocks)

    @property
    def tie_layers(self) -> bool:
        return (self.n_layer == 1) or (
            all(
                [
                    (a.encoder is b.encoder) and (a.decoder is b.decoder)
                    for a, b in product(self.blocks[:1], self.blocks[1:])
                ]
            )
        )

    def register_loss_func(self, func: LossFunction) -> None:
        if not isinstance(func, LossFunction):
            raise ValueError(
                "Expect a `LossFunction` object, "
                f"but receive a `{type(func).__class__.__name__}` object."
            )
        if hasattr(self, "_loss_hook_handle"):
            self._loss_hook_handle.remove()
            del self._loss_hook_handle
            del self._loss_func
        self._loss_func = func
        self._loss_hook_handle = self.register_forward_hook(
            self._loss_func.loss_hook
        )

    def _step(
        self,
        data: Tensor,
        feats: Optional[Tensor],
        mask: Optional[BoolTensor],
    ) -> Tuple[Tensor, Tensor]:
        cumsum = pt.tensor(0, dtype=data.dtype, device=data.device)
        residue = []
        forecast = []
        for block in self.blocks:
            interp, extrap = block(data, feats, mask)
            data = data - interp
            cumsum = cumsum + extrap
            residue.append(data)
            forecast.append(cumsum)
        residue = pt.stack(residue, dim=2)
        forecast = pt.stack(forecast, dim=2)
        return residue, forecast

    def _rollout(
        self,
        data: Tensor,
        feats: Optional[Tensor],
        mask: Optional[BoolTensor],
        total_steps: Optional[int] = None,
    ) -> Tuple[Tensor, Tensor]:
        if feats is not None:
            prefix_feats = feats[:, : data.size(1)]
            n_steps = feats.size(1) - data.size(1) + 1
            if total_steps is None:
                total_steps = n_steps
            else:
                total_steps = min(total_steps, n_steps)
        else:
            if (total_steps is None) and (not self.training):
                raise ValueError("Total decoding steps must be provided.")
            prefix_feats = None

        residue, forecast = self._step(data, prefix_feats, mask)
        if not self.training:
            preds = forecast
            forecast = pt.narrow(forecast, dim=1, start=0, length=0)
            for step in range(total_steps):
                prev = pt.narrow(preds, dim=1, start=-1, length=1)
                forecast = pt.cat([forecast, prev], dim=1)
                prev = prev[..., -1, :]
                data = pt.cat([data, prev.detach()], dim=1)
                if feats is not None:
                    prefix_feats = pt.narrow(
                        feats, dim=1, start=0, length=data.size(1)
                    )
                if mask is not None:
                    mask = pt.cat(
                        [
                            mask,
                            pt.zeros_like(
                                pt.narrow(mask, dim=1, start=0, length=1)
                            ),
                        ],
                        dim=1,
                    )
                _, preds = self._step(data, prefix_feats, mask)

        return residue, forecast

    def forward(
        self,
        data: Tensor,
        feats: Optional[Tensor],
        nan_mask: Optional[BoolTensor],
        length: Optional[LongTensor],
    ) -> Tuple[
        Tensor, Tensor, Tensor, Optional[BoolTensor], Optional[BoolTensor]
    ]:
        batch_size = data.size(0)
        full_length = data.size(1)

        # merge padding mask and missing mask
        if length is None:
            pad_mask = None
        else:
            # in our impl, padings are on left side
            pad_mask = pt.gt(
                pt.arange(full_length, 0, -1).to(length).unsqueeze(dim=0),
                length.unsqueeze(dim=1),
            )
        if (pad_mask is None) and (nan_mask is None):
            mask = None
        else:
            if pad_mask is None:
                pad_mask = pt.zeros_like(nan_mask)
            if nan_mask is None:
                nan_mask = pt.zeros_like(pad_mask)
            mask = pt.logical_or(pad_mask, nan_mask)

        # compute acutal horizons
        if isinstance(self.horizon, int):
            horizon = data.new_ones(batch_size, dtype=pt.long) * self.horizon
        else:
            if length is None:
                horizon = data.new_ones(batch_size, dtype=pt.long) * int(
                    full_length * self.horizon
                )
            else:
                horizon = length.to(pt.float).mul(self.horizon).to(pt.long)
        max_horizon = horizon.max().item()

        # split data into condition (input) and (forecasting) target
        if self.training:
            target = data[:, full_length - max_horizon : full_length]
            if mask is not None:
                bc_mask = mask
                fc_mask = pt.gt(
                    pt.arange(max_horizon, 0, -1).to(horizon).unsqueeze(dim=0),
                    horizon.unsqueeze(dim=1),
                )
                fc_mask = pt.logical_or(
                    fc_mask, mask[:, full_length - max_horizon : full_length]
                )
            else:
                bc_mask = fc_mask = None
            residue, forecast = self._rollout(data, feats, bc_mask)
            forecast = forecast[:, -(max_horizon + 1) : -1]
        else:
            if length is not None:
                # move part of paddings from left to right
                # pad_size = max_horizon - horizon
                start_index = pt.unsqueeze(max_horizon - horizon, dim=1)
                selected_index = (
                    pt.arange(full_length).to(horizon).unsqueeze(dim=0)
                    + start_index
                )
                selected_index = pt.where(
                    selected_index.lt(full_length),
                    selected_index,
                    selected_index - full_length,
                )
                data = pt.gather(
                    data,
                    dim=1,
                    index=selected_index.unsqueeze(dim=2).expand_as(data),
                )
                if feats is not None:
                    feats = pt.gather(
                        feats,
                        dim=1,
                        index=selected_index.unsqueeze(dim=2).expand_as(feats),
                    )
                mask = pt.gather(
                    mask,
                    dim=1,
                    index=selected_index,
                )

            # split condition and target
            cond_length = full_length - max_horizon
            data, target = pt.split(data, [cond_length, max_horizon], dim=1)
            # features are not split for predictions
            if mask is not None:
                bc_mask, fc_mask = pt.split(
                    mask, [cond_length, max_horizon], dim=1
                )
            else:
                bc_mask = fc_mask = None
            residue, forecast = self._rollout(
                data, feats, bc_mask, max_horizon
            )

        return residue, forecast, target, bc_mask, fc_mask

    @classmethod
    def from_configs(
        cls,
        d_data: int,
        d_feats: int,
        d_hidden: int,
        n_layer: int,
        horizon: Union[int, float],
        window_size: List[int],
        n_head: int = 4,
        n_enc_layer: int = 2,
        n_dec_layer: int = 2,
        symmetric: bool = False,
        share_values: bool = False,
        tie_casts: bool = False,
        tie_layers: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
        norm: str = "instance",
        add_dist: bool = False,
        normalize_input: bool = True,
        rescale_output: bool = True,
        layerwise_loss: bool = True,
    ):
        kernel = ExpKernel(
            d_hidden,
            n_head,
            n_enc_layer,
            n_dec_layer,
            symmetric=symmetric,
            share_values=share_values,
            dropout=dropout,
            temperature=temperature,
            norm=norm,
            add_dist=add_dist,
        )
        encoder = EncoderModule(
            d_data,
            d_feats,
            d_hidden,
            kernel.d_head if share_values else d_hidden,
            *window_size,
            tie_casts=tie_casts,
        )
        decoder = DecoderModule(d_data, d_hidden)
        if tie_layers:
            encoders = [encoder] * n_layer
            decoders = [decoder] * n_layer
            kernels = [kernel] * n_layer
        else:
            encoders = [deepcopy(encoder) for _ in range(n_layer)]
            decoders = [deepcopy(decoder) for _ in range(n_layer)]
            kernels = [deepcopy(kernel) for _ in range(n_layer)]

        return cls(
            encoders,
            kernels,
            decoders,
            horizon=horizon,
            normalize_input=normalize_input,
            rescale_output=rescale_output,
            layerwise_loss=layerwise_loss,
        )

    def create_twin_estimator(
        self,
        d_data: int,
        d_feats: int,
        **distinct_configs,
    ):
        kernels = [block.kernel for block in self.blocks]
        d_value = (
            kernels[0].d_head
            if kernels[0].share_values
            else kernels[0].d_hidden
        )
        d_hidden = kernels[0].d_hidden
        window_size = distinct_configs.get(
            "window_size", self.blocks[0].encoder.window_size
        )
        tie_casts = distinct_configs.get(
            "tie_casts", self.blocks[0].encoder.tie_casts
        )
        estimator_configs = dict(
            horizon=distinct_configs.get("horizon", self.horizon),
            loss_func=distinct_configs.get(
                "loss_func", type(self._loss_func)()
            ),
            normalize_input=distinct_configs.get(
                "normalize_input", self.normalize_input
            ),
            rescale_output=distinct_configs.get(
                "rescale_output", self.rescale_output
            ),
            layerwise_loss=distinct_configs.get(
                "layerwise_loss", self.layerwise_loss
            ),
        )

        encoder = EncoderModule(
            d_data,
            d_feats,
            d_hidden,
            d_value,
            *window_size,
            tie_casts=tie_casts,
        )
        decoder = DecoderModule(
            d_data,
            d_hidden,
        )
        if self.tie_layers:
            encoders = [encoder] * self.n_layer
            decoders = [decoder] * self.n_layer
        else:
            encoders = [deepcopy(encoder) for _ in range(self.n_layer)]
            decoders = [deepcopy(decoder) for _ in range(self.n_layer)]
        return type(self)(encoders, kernels, decoders, **estimator_configs)


class AdversarialEstimator(AttentionEstimator):
    def __init__(
        self,
        *lists_of_modules: List[nn.Module],
        horizon: Union[int, float],
        loss_func: LossFunction = MSELoss(),
        normalize_input: bool = True,
        rescale_output: bool = True,
        layerwise_loss: bool = True,
    ) -> None:
        super(AdversarialEstimator, self).__init__(
            *lists_of_modules,
            horizon=horizon,
            loss_func=loss_func,
            normalize_input=normalize_input,
            rescale_output=rescale_output,
            layerwise_loss=layerwise_loss,
        )
        self._generative = True

    def _create_modules(self, *modules):
        self.blocks = nn.ModuleList(
            [AdversarialBlock(*ms) for ms in zip(*modules)]
        )

    def generative(self) -> None:
        self._generative = True
        for block in self.blocks:
            block.generative()

    def discriminative(self) -> None:
        self._generative = False
        for block in self.blocks:
            block.discriminative()

    def generative_named_parameters(self) -> Iterator[Tuple[str, Parameter]]:
        for i, block in enumerate(self.blocks):
            for n, p in block.encoder.named_parameters():
                yield f"blocks.{i}.encoder.{n}", p
            for n, p in block.kernel.named_parameters():
                yield f"blocks.{i}.kernel.{n}", p
            for n, p in block.decoder.named_parameters():
                yield f"blocks.{i}.decoder.{n}", p

    def discriminative_named_parameters(
        self,
    ) -> Iterator[Tuple[str, Parameter]]:
        for i, block in enumerate(self.blocks):
            for n, p in block.disc.named_parameters():
                yield f"blocks.{i}.disc.{n}", p

    def generative_parameters(self) -> Iterator[Parameter]:
        for n, p in self.generative_named_parameters():
            yield p

    def discriminative_parameters(self) -> Iterator[Parameter]:
        for n, p in self.discriminative_named_parameters():
            yield p

    @property
    def prob_domain(self) -> Tensor:
        return pt.stack([block.prob_domain for block in self.blocks], dim=2)

    @classmethod
    def from_configs(
        cls,
        d_data: int,
        d_feats: int,
        d_hidden: int,
        n_layer: int,
        horizon: Union[int, float],
        window_size: List[int],
        n_head: int = 4,
        n_enc_layer: int = 2,
        n_dec_layer: int = 2,
        n_disc_layer: int = 2,
        symmetric: bool = False,
        share_values: bool = True,
        tie_casts: bool = False,
        tie_layers: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
        norm: str = "instance",
        add_dist: bool = False,
        normalize_input: bool = True,
        rescale_output: bool = True,
        layerwise_loss: bool = True,
    ):
        kernel = ExpKernel(
            d_hidden,
            n_head,
            n_enc_layer,
            n_dec_layer,
            symmetric=symmetric,
            share_values=share_values,
            dropout=dropout,
            temperature=temperature,
            norm=norm,
            add_dist=add_dist,
        )
        disc = SimpleDiscriminator(
            d_model=d_hidden,
            n_layer=n_disc_layer,
        )
        encoder = EncoderModule(
            d_data,
            d_feats,
            d_hidden,
            kernel.d_head if share_values else d_hidden,
            *window_size,
            tie_casts=tie_casts,
        )
        decoder = DecoderModule(
            d_data,
            d_hidden,
        )
        if tie_layers:
            encoders = [encoder] * n_layer
            decoders = [decoder] * n_layer
            kernels = [kernel] * n_layer
            discs = [disc] * n_layer
        else:
            encoders = [deepcopy(encoder) for _ in range(n_layer)]
            decoders = [deepcopy(decoder) for _ in range(n_layer)]
            kernels = [deepcopy(kernel) for _ in range(n_layer)]
            discs = [deepcopy(disc) for _ in range(n_layer)]

        return cls(
            encoders,
            kernels,
            decoders,
            discs,
            horizon=horizon,
            normalize_input=normalize_input,
            rescale_output=rescale_output,
            layerwise_loss=layerwise_loss,
        )

    @classmethod
    def from_base(
        cls,
        base: AttentionEstimator,
        n_disc_layer: int = 2,
    ):
        kernels = [deepcopy(block.kernel) for block in base.blocks]
        encoders = [deepcopy(block.encoder) for block in base.blocks]
        decoders = [deepcopy(block.decoder) for block in base.blocks]

        disc = SimpleDiscriminator(
            d_model=kernels[0].d_hidden, n_layer=n_disc_layer
        )
        n_layer = base.n_layer
        if base.tie_layers:
            discs = [disc] * n_layer
        else:
            discs = [deepcopy(disc) for _ in range(n_layer)]
        return cls(
            encoders,
            kernels,
            decoders,
            discs,
            horizon=base.horizon,
            loss_func=type(base._loss_func)(),
            normalize_input=base.normalize_input,
            rescale_output=base.rescale_output,
            layerwise_loss=base.layerwise_loss,
        )

    def create_twin_estimator(
        self,
        d_data: int,
        d_feats: int,
        **distinct_configs,
    ):
        kernels = [block.kernel for block in self.blocks]
        discs = [block.disc for block in self.blocks]
        d_value = (
            kernels[0].d_head
            if kernels[0].share_values
            else kernels[0].d_hidden
        )
        d_hidden = kernels[0].d_hidden
        window_size = distinct_configs.get(
            "window_size", self.blocks[0].encoder.window_size
        )
        tie_casts = distinct_configs.get(
            "tie_casts", self.blocks[0].encoder.tie_casts
        )
        estimator_configs = dict(
            horizon=distinct_configs.get("horizon", self.horizon),
            loss_func=distinct_configs.get(
                "loss_func", type(self._loss_func)()
            ),
            normalize_input=distinct_configs.get(
                "normalize_input", self.normalize_input
            ),
            rescale_output=distinct_configs.get(
                "rescale_output", self.rescale_output
            ),
            layerwise_loss=distinct_configs.get(
                "layerwise_loss", self.layerwise_loss
            ),
        )

        encoder = EncoderModule(
            d_data,
            d_feats,
            d_hidden,
            d_value,
            *window_size,
            tie_casts=tie_casts,
        )
        decoder = DecoderModule(
            d_data,
            d_hidden,
        )
        if self.tie_layers:
            encoders = [encoder] * self.n_layer
            decoders = [decoder] * self.n_layer
        else:
            encoders = [deepcopy(encoder) for _ in range(self.n_layer)]
            decoders = [deepcopy(decoder) for _ in range(self.n_layer)]
        return type(self)(
            encoders, kernels, decoders, discs, **estimator_configs
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import torch as pt
from torch import Tensor, BoolTensor
from torch.nn.functional import l1_loss, mse_loss
from torch.distributions import Distribution, Normal as Gaussian


def quantile_error(preds: Tensor, target: Tensor, percentage: float) -> Tensor:
    diff = target - preds
    weight = pt.where(
        condition=diff > 0,
        input=diff.new_tensor(percentage),
        other=diff.new_tensor(percentage - 1),
    )
    return diff * weight


def quantile_percent_error(
    preds: Tensor, target: Tensor, percentage: float
) -> Tensor:
    qe = quantile_error(preds, target, percentage)
    qpe = qe * 2 / target
    qpe = qpe.masked_fill_(target <= 0, 0.0)
    return qpe


def absolute_error(preds: Tensor, target: Tensor) -> Tensor:
    return quantile_error(preds, target, 0.5) * 2


def absolute_percent_error(preds: Tensor, target: Tensor) -> Tensor:
    return quantile_percent_error(preds, target, 0.5)


def root_square_error(preds: Tensor, target: Tensor) -> Tensor:
    return mse_loss(preds, target, reduction="none").sqrt()


def root_square_percent_error(preds: Tensor, target: Tensor) -> Tensor:
    se = root_square_error(preds, target)
    rspe = se / target.abs()
    return rspe


def tweedie_loss(p: float):
    def tweedie_ll(preds: Tensor, target: Tensor):
        return -target * pt.pow(preds, 1 - p) / (1 - p) + pt.pow(
            preds, 2 - p
        ) / (2 - p)

    return tweedie_ll



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations
from typing import TYPE_CHECKING, Optional, List, Dict
from pathlib import Path
import math

from tqdm import tqdm, trange
from torch.utils.tensorboard import SummaryWriter
from torch.nn.utils import clip_grad_norm_

if TYPE_CHECKING:
    from .trainer import Trainer


class Callback(object):
    def __init__(self, trainer: Optional[Trainer] = None):
        if trainer is not None:
            self.set_trainer(trainer)
        else:
            self.trainer = None

    def set_trainer(self, trainer: Trainer):
        self.trainer = trainer

    def on_train_begin(self, **kwargs):
        pass

    def on_train_end(self, **kwargs):
        pass

    def on_epoch_begin(self, epoch, **kwargs):
        pass

    def on_epoch_end(self, epoch, **kwargs):
        pass

    def on_batch_begin(self, batch, **kwargs):
        pass

    def on_batch_end(self, batch, **kwargs):
        pass

    def on_valid_begin(self, epoch, **kwargs):
        pass

    def on_valid_end(self, epoch, **kwargs):
        pass

    def on_valid_batch_begin(self, batch, **kwargs):
        pass

    def on_valid_batch_end(self, batch, **kwargs):
        pass


class CallbackList(Callback):
    def __init__(self, *callbacks, trainer: Optional[Trainer] = None):
        self.callbacks = list(callbacks)
        super(CallbackList, self).__init__(trainer)

    def set_trainer(self, trainer: Trainer):
        self.trainer = trainer
        for callback in self.callbacks:
            callback.set_trainer(trainer)

    def append(self, callback: Callback):
        if not isinstance(callback, Callback):
            raise ValueError("The given object is not a callback.")
        if isinstance(callback, CallbackList):
            raise ValueError("Cannot add a callback list to a callback list")
        callback.set_trainer(self.trainer)
        self.callbacks.append(callback)

    def __len__(self):
        return len(self.callbacks)

    def __getitem__(self, index):
        return self.callbacks[index]

    def on_train_begin(self, **kwargs):
        for callback in self.callbacks:
            callback.on_train_begin(**kwargs)

    def on_train_end(self, **kwargs):
        for callback in self.callbacks:
            callback.on_train_end(**kwargs)

    def on_epoch_begin(self, epoch, **kwargs):
        for callback in self.callbacks:
            callback.on_epoch_begin(epoch, **kwargs)

    def on_epoch_end(self, epoch, **kwargs):
        for callback in self.callbacks:
            callback.on_epoch_end(epoch, **kwargs)

    def on_batch_begin(self, batch, **kwargs):
        for callback in self.callbacks:
            callback.on_batch_begin(batch, **kwargs)

    def on_batch_end(self, batch, **kwargs):
        for callback in self.callbacks:
            callback.on_batch_end(batch, **kwargs)

    def on_valid_begin(self, epoch, **kwargs):
        for callback in self.callbacks:
            callback.on_valid_begin(epoch, **kwargs)

    def on_valid_end(self, epoch, **kwargs):
        for callback in self.callbacks:
            callback.on_valid_end(epoch, **kwargs)

    def on_valid_batch_begin(self, batch, **kwargs):
        for callback in self.callbacks:
            callback.on_valid_batch_begin(batch, **kwargs)

    def on_valid_batch_end(self, batch, **kwargs):
        for callback in self.callbacks:
            callback.on_valid_batch_end(batch, **kwargs)


class Optimization(Callback):
    def __init__(
        self,
        grad_cum_step: int = 1,
        max_grad_norm: Optional[float] = None,
        lr_decay_freq: str = "batch",
    ):
        super(Optimization, self).__init__()
        if lr_decay_freq not in ["batch", "epoch"]:
            raise ValueError(
                f'learning rate decay frequency must be either "batch" or "epoch"'
            )
        self.grad_cum_step = grad_cum_step
        self.lr_decay_freq = lr_decay_freq
        self.max_grad_norm = max_grad_norm

    def on_train_begin(self, **kwargs):
        self.trainer.optimizer.zero_grad()
        self.grad_cum_counter = self.grad_cum_step

    def on_batch_end(self, batch, **kwargs):
        self.grad_cum_counter -= 1
        if self.grad_cum_counter == 0:
            if self.max_grad_norm is not None:
                grad_norm = clip_grad_norm_(
                    self.trainer.model.parameters(), self.max_grad_norm
                )
                if "grad_norm" in self.trainer.metrics.train:
                    self.trainer.metrics.train.grad_norm.update(grad_norm)
            self.trainer.optimizer.step()
            self.trainer.global_step += 1
            if (
                self.lr_decay_freq == "batch"
                and self.trainer.scheduler is not None
            ):
                self.trainer.scheduler.step()
            self.trainer.optimizer.zero_grad()
            self.grad_cum_counter = self.grad_cum_step
        else:
            pass

    def on_epoch_end(self, epoch, **kwargs):
        if (
            self.lr_decay_freq == "epoch"
            and self.trainer.scheduler is not None
        ):
            self.trainer.scheduler.step()


class DebugStopper(Callback):
    def __init__(self, debug_stop_batch: int = 1, debug_stop_epoch: int = 1):
        super(DebugStopper, self).__init__()
        self.debug_stop_batch = debug_stop_batch
        self.debug_stop_epoch = debug_stop_epoch

    def on_train_begin(self, **kwargs):
        self.trainer._signal_exit = False
        self.trainer._signal_break = False
        self.trainer._signal_valid_break = False

    def on_batch_end(self, batch, **kwargs):
        if self.trainer.debug:
            if batch >= self.debug_stop_batch - 1:
                self.trainer._signal_break = True
        else:
            self.trainer._signal_break = False

    def on_epoch_end(self, epoch, **kwargs):
        if self.trainer.debug:
            if epoch >= self.debug_stop_epoch - 1:
                self.trainer._signal_exit = True
        else:
            self.trainer._signal_exit = False

    def on_valid_batch_end(self, batch, **kwargs):
        if self.trainer.debug:
            if batch >= self.debug_stop_batch - 1:
                self.trainer._signal_valid_break = True
        else:
            self.trainer._signal_valid_break = False


class EarlyStopping(Callback):
    def __init__(
        self,
        patience: int,
        monitor: str = "valid/loss",
        minimize: bool = True,
        threshold: float = 0.0,
    ):
        super(EarlyStopping, self).__init__()
        self.patience = patience
        self.monitor = monitor
        self.minimize = minimize
        self.threshold = threshold

    def on_train_begin(self, **kwargs):
        self.wait = 0
        self._stopped_epoch = 0
        self.best = float("inf")
        if not self.minimize:
            self.best = -self.best

    def on_epoch_end(self, epoch, **kwargs):
        current_value = self.trainer.metrics[self.monitor].value
        diff = current_value - self.best
        better = (
            (diff < -self.threshold)
            if self.minimize
            else (diff > self.threshold)
        )
        if better:
            self.best = current_value
            self.wait = 0
        else:
            self.wait += 1
            if self.wait == self.patience:
                self.trainer._signal_exit = True
                self._stopped_epoch = epoch + 1

    def on_train_end(self, **kwargs):
        if self._stopped_epoch > 0:
            print(f"Early stopping at epoch {self._stopped_epoch:02d}")


class Checkpoint(Callback):
    def __init__(
        self,
        monitor: str = "valid/loss",
        minimize: bool = True,
        threshold: float = 0.0,
        keep_epoch_dump: bool = True,
    ):
        super(Checkpoint, self).__init__()
        self.monitor = monitor
        self.minimize = minimize
        self.threshold = threshold
        self.keep_epoch_dump = keep_epoch_dump

    def on_train_begin(self, **kwargs):
        self.best = float("inf") if self.minimize else float("-inf")

    def on_epoch_end(self, epoch, **kwargs):
        current_value = self.trainer.metrics[self.monitor].value
        diff = current_value - self.best
        better = (
            (diff < -self.threshold)
            if self.minimize
            else (diff > self.threshold)
        )
        if better:
            self.best = current_value
            self.trainer.dump(tag="best")
        if self.keep_epoch_dump:
            self.trainer.dump(tag=f"epoch{epoch+1:02d}")


class ProgressBar(Callback):
    def __init__(
        self,
        batch_monitors: Optional[List[str]] = None,
        epoch_monitors: Optional[List[str]] = None,
    ):
        super(ProgressBar, self).__init__()
        self.batch_monitors = batch_monitors or []
        self.epoch_monitors = epoch_monitors or []

    def on_epoch_begin(self, epoch: int, **kwargs):
        if self.trainer.nb_epoch is None:
            n_batch = int(
                math.ceil(
                    self.trainer.dataset.train_size / self.trainer.batch_size
                )
            )
        else:
            n_batch = self.trainer.nb_epoch
        device = (
            "cpu"
            if self.trainer.cuda_device < 0
            else f"cuda:{self.trainer.cuda_device}"
        )
        self.prog_bar = tqdm(
            desc=f"({device}) {epoch+1}/{self.trainer.n_epochs}",
            total=n_batch,
            unit="batch",
        )

    def on_batch_end(self, batch, **kwargs):
        self.prog_bar.update(1)
        postfix = {}
        for k in self.batch_monitors:
            meter = self.trainer.metrics.get(k)
            if meter is None:
                v = kwargs[k]
            else:
                v = meter.value
            postfix[k] = v
        self.prog_bar.set_postfix(postfix)

    def on_valid_begin(self, epoch, **kwargs):
        n_batch = int(
            math.ceil(
                self.trainer.dataset.valid_size / self.trainer.eval_batch_size
            )
        )
        device = (
            "cpu"
            if self.trainer.cuda_device < 0
            else f"cuda:{self.trainer.cuda_device}"
        )
        self.sub_prog_bar = tqdm(
            desc=f"    ({device}) validating epoch {epoch+1}",
            total=n_batch,
            unit="batch",
            position=1,
            leave=False,
        )

    def on_valid_batch_end(self, batch, **kwargs):
        self.sub_prog_bar.update(1)

    def on_valid_end(self, epoch, **kwargs):
        self.sub_prog_bar.close()

    def on_epoch_end(self, epoch, **kwargs):
        postfix = {}
        for k in self.epoch_monitors:
            meter = self.trainer.metrics.get(k)
            if meter is None:
                v = kwargs[k]
            else:
                v = meter.value
            postfix[k] = v
        self.prog_bar.set_postfix(postfix)
        self.prog_bar.close()


class Tensorboard(Callback):
    def __init__(
        self,
        batch_monitors: Optional[List[str]] = None,
        epoch_monitors: Optional[List[str]] = None,
        hparams: Optional[Dict] = None,
        metrics: Optional[List[str]] = None,
        comment: str = "",
    ):
        super(Tensorboard, self).__init__()
        self.comment = comment
        self.batch_monitors = batch_monitors or []
        self.epoch_monitors = epoch_monitors or []
        self.hparams = hparams
        self.metrics = metrics

    def on_train_begin(self, **kwargs):
        self.writer = SummaryWriter(self.trainer.log_dir, self.comment)

    def on_batch_end(self, batch, **kwargs):
        for k in self.batch_monitors:
            v = kwargs.get(k)
            if v is None:
                v = self.trainer.metrics[k].cache
            self.writer.add_scalar(k, v, global_step=self.trainer.global_step)

    def on_epoch_end(self, epoch, **kwargs):
        for k in self.epoch_monitors:
            v = kwargs.get(k)
            if v is None:
                v = self.trainer.metrics[k].value
            self.writer.add_scalar(k, v, global_step=self.trainer.global_step)
        weights = kwargs.get("weights", {})
        for k, v in weights.items():
            self.writer.add_histogram(
                f"weights/{k}", v, global_step=self.trainer.global_step
            )
        grads = kwargs.get("grads", {})
        for k, v in grads.items():
            self.writer.add_histogram(
                f"grads/{k}", v, global_step=self.trainer.global_step
            )

    def on_train_end(self, **kwargs):
        if self.hparams is not None:
            if self.metrics is None:
                metrics = self.metrics.valid.best
            else:
                metrics = {
                    k: self.trainer.metrics[k].best for k in self.metrics
                }
            self.writer.add_hparams(self.hparams, metrics)
        self.writer.close()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Callable, Optional, List, Tuple, Dict, Union
from functools import partial
from itertools import chain
from pathlib import Path

from torch import Tensor, BoolTensor
from torch.optim import AdamW
import numpy as np

from tslib.dataset import MetaDataset
from tslib.engine import Trainer, Evaluator, HyperOptManager
from tslib.engine.distributed import (
    reduce_value,
    is_main_process,
)
from tslib.engine.callback import *
from tslib.metrics import (
    NumericalAverageMeter,
    BatchAverageMeter,
    MeanDeviationMeter,
    MeterDict,
)
from ..estimator import AttentionEstimator


class AttentionTrainer(Trainer):
    def _batch_loop(self, epoch: int):
        for batch, data in enumerate(self.train_loader):
            self.callbacks.on_batch_begin(batch)
            self._train(*data)
            self.callbacks.on_batch_end(
                batch,
            )
            if self._signal_break:
                break

    def _train(self, *data):
        loss = self.model(*data)
        loss.mean().backward()
        loss = reduce_value(loss, dst=None)
        self.metrics.train.loss.update(loss)
        self.metrics.train.bc_loss.update(self.model.bc_loss)
        self.metrics.train.fc_loss.update(self.model.fc_loss)

    def _validate(self, *data):
        _ = self.model(*data)
        self.metrics.valid.bc_loss.update(self.model.bc_loss)
        self.metrics.valid.fc_loss.update(self.model.fc_loss)
        self.metrics.valid.fc_ND.update(
            self.model.fc_loss, self.model.denominator
        )

    @classmethod
    def from_configs(
        cls,
        dataset: MetaDataset,
        model: AttentionEstimator,
        log_dir: Union[Path, str],
        cuda_device: int,
        n_epochs: int,
        nb_epoch: Optional[int] = None,
        batch_size: Union[int, Tuple[int, int]] = 100,
        max_grad_norm: float = 1e0,
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        weight_decay: float = 0.0,
        amsgrad: bool = False,
        debug: bool = False,
        **kwargs,
    ):
        optimizer = partial(
            AdamW,
            lr=lr,
            betas=betas,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
        )

        meters = MeterDict(
            meterdicts={
                "train": MeterDict(
                    meters=dict(
                        loss=BatchAverageMeter(),
                        grad_norm=NumericalAverageMeter(),
                        fc_loss=BatchAverageMeter(),
                        bc_loss=BatchAverageMeter(),
                    )
                ),
                "valid": MeterDict(
                    meters=dict(
                        fc_loss=BatchAverageMeter(),
                        bc_loss=BatchAverageMeter(),
                        fc_ND=MeanDeviationMeter(),
                    )
                ),
                "test": MeterDict(
                    meters=dict(
                        fc_loss=BatchAverageMeter(),
                        bc_loss=BatchAverageMeter(),
                        fc_ND=MeanDeviationMeter(),
                    )
                ),
            }
        )

        callbacks = [
            Optimization(max_grad_norm=max_grad_norm),
            ProgressBar(
                batch_monitors=["train/loss", "train/grad_norm"],
                epoch_monitors=[
                    "train/fc_loss",
                    "valid/fc_loss",
                    "valid/fc_ND",
                ],
            ),
            Checkpoint(monitor="valid/fc_loss"),
            Tensorboard(
                batch_monitors=[
                    "train/loss",
                    "train/bc_loss",
                    "train/fc_loss",
                ],
                epoch_monitors=[
                    "valid/bc_loss",
                    "valid/fc_loss",
                    "valid/fc_ND",
                ],
            ),
        ]

        return cls(
            dataset,
            model,
            optimizer,
            None,
            metrics=meters,
            callbacks=callbacks,
            log_dir=log_dir,
            n_epochs=n_epochs,
            nb_epoch=nb_epoch,
            batch_size=batch_size,
            cuda_device=cuda_device,
            debug=debug,
        )


class AttentionEvaluator(Evaluator):
    def _test(self, *data):
        _ = self.model(*data)
        bc_loss = self.model.bc_loss
        fc_loss = self.model.fc_loss
        self.metrics.test.bc_loss.update(self.model.bc_loss)
        self.metrics.test.fc_loss.update(self.model.fc_loss)
        self.metrics.test.fc_ND.update(
            self.model.fc_loss, self.model.denominator
        )

    def _predict(self, *data) -> np.ndarray:
        _ = self.model(*data)
        truth = data[0].cpu().numpy()
        preds = self.model.forecast[..., 0].cpu().numpy()
        dummy = np.zeros(truth.shape[:-1] + preds.shape[-1:]) + np.nan
        preds = np.concatenate([dummy[:, : -preds.shape[1]], preds], axis=1)
        curve = np.concatenate([truth, preds], axis=-1)
        return curve

    @classmethod
    def from_trainer(
        cls,
        trainer: AttentionTrainer,
        model_tag: str = "best",
    ):
        return super(AttentionEvaluator, cls).from_trainer(trainer, model_tag)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict

import numpy as np
import pandas as pd

from src.gluonts.dataset.common import DataEntry, Dataset
from src.gluonts.dataset.field_names import FieldName


def make_predictions(
    pred_input: Dataset, predictor, forecasts_at_all_levels: bool = False
):
    predictor.prediction_net.return_forecasts_at_all_levels = (
        forecasts_at_all_levels
    )
    forecast_it = predictor.predict(pred_input)
    return forecast_it


def truncate_target(
    dataset: Dataset, prediction_length: int, lead_time: int = 0
):
    def _truncate(data: DataEntry):
        data = data.copy()
        target = data["target"]
        assert (
            target.shape[-1] >= prediction_length
        )  # handles multivariate case (target_dim, history_length)
        data["target"] = target[..., : -prediction_length - lead_time]
        return data

    return map(_truncate, dataset)


def to_dataframe_it(test):
    def period_index(entry: DataEntry, freq=None) -> pd.DatetimeIndex:
        if freq is None:
            freq = entry[FieldName.START].freq

        return pd.period_range(
            start=entry[FieldName.START],
            periods=entry[FieldName.TARGET].shape[-1],
            freq=freq,
        )

    def _to_dataframe(entry: Dict):
        return pd.DataFrame(entry["target"], index=period_index(entry))

    return map(_to_dataframe, test)


def crps(
    forecast,
    actuals,
    levels: np.ndarray = (np.arange(10) / 10.0)[1:],
    weighted: bool = True,
):
    quantiles = np.array([forecast.quantile(level) for level in levels])

    levels = np.expand_dims(levels, axis=-1)

    crps = np.where(
        actuals >= quantiles,
        levels * (actuals - quantiles),
        (1 - levels) * (quantiles - actuals),
    )

    if weighted:
        return crps.sum() / actuals.sum()
    else:
        return crps.sum()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional
import os

import torch as pt
from torch import Tensor
from torch import distributed as dist


def is_distributed():
    return dist.is_available() and dist.is_initialized()


def get_world_size():
    if not is_distributed():
        return 1
    return dist.get_world_size()


def get_rank():
    if not is_distributed():
        return 0
    return dist.get_rank()


def is_main_process():
    return get_rank() == 0


def synchronize():
    if get_world_size() == 1:
        return
    dist.barrier()


def reduce_value(value: Tensor, dst: Optional[int] = 0, average: bool = True):
    world_size = get_world_size()
    if world_size < 2:
        return value
    if dst is None:
        dist.all_reduce(value)
        if average:
            value = value / world_size
    else:
        dist.reduce(value, dst)
        if (get_rank() == dst) and average:
            value = value / world_size
    return value


def init_distributed(
    cuda: bool,
    local_rank: int,
    node_rank: int = 0,
    n_nodes: int = 1,
    n_procs: int = 0,
    master_addr: str = "localhost",
    master_port: int = 60066,
):
    if not dist.is_available():
        return
    cuda = pt.cuda.is_available() and cuda
    if cuda:
        if n_procs <= 0:
            n_procs = pt.cuda.device_count()
        else:
            assert (
                n_procs <= pt.cuda.device_count()
            ), f"# processes per node exceeds # available GPUs"
    else:
        if n_procs <= 0:
            n_procs = os.cpu_count()
        else:
            assert (
                n_procs <= os.cpu_count()
            ), f"# processes per node exceeds # available CPUs"
    assert (
        0 <= node_rank < n_nodes
    ), f"rank {node_rank} is not one of [0,...,{n_nodes-1}]"
    assert (
        0 <= local_rank < n_procs
    ), f"local_rank {local_rank} is not one of [0, ..., {n_procs-1}]"

    world_size = n_nodes * n_procs
    if world_size == 1:
        return
    os.environ["MASTER_ADDR"] = master_addr
    os.environ["MASTER_PORT"] = str(master_port)
    os.environ["WORLD_SIZE"] = str(world_size)
    rank = node_rank * n_procs + local_rank
    dist.init_process_group("nccl" if cuda else "gloo", rank=rank)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from torch.distributions import Distribution

from src.gluonts.core.component import validated
from pts.model import weighted_average

from .tft_modules import (
    FeatureProjector,
    FeatureEmbedder,
    VariableSelectionNetwork,
    GatedResidualNetwork,
    TemporalFusionEncoder,
    TemporalFusionDecoder,
)
from .tft_output import QuantileOutput


class TemporalFusionTransformerNetwork(nn.Module):
    @validated()
    def __init__(
        self,
        context_length: int,
        prediction_length: int,
        variable_dim: int,
        embed_dim: int,
        num_heads: int,
        num_outputs: int,
        d_past_feat_dynamic_real: List[int],
        c_past_feat_dynamic_cat: List[int],
        d_feat_dynamic_real: List[int],
        c_feat_dynamic_cat: List[int],
        d_feat_static_real: List[int],
        c_feat_static_cat: List[int],
        dropout: float = 0.0,
    ):
        super().__init__()

        self.context_length = context_length
        self.prediction_length = prediction_length
        self.normalize_eps = 1e-5

        self.target_proj = nn.Linear(in_features=1, out_features=variable_dim)

        if d_past_feat_dynamic_real:
            self.past_feat_dynamic_proj = FeatureProjector(
                feature_dims=d_past_feat_dynamic_real,
                embedding_dims=[variable_dim] * len(d_past_feat_dynamic_real),
            )
        else:
            self.past_feat_dynamic_proj = None

        if c_past_feat_dynamic_cat:
            self.past_feat_dynamic_embed = FeatureEmbedder(
                cardinalities=c_past_feat_dynamic_cat,
                embedding_dims=[variable_dim] * len(c_past_feat_dynamic_cat),
            )
        else:
            self.past_feat_dynamic_embed = None

        if d_feat_dynamic_real:
            self.feat_dynamic_proj = FeatureProjector(
                feature_dims=d_feat_dynamic_real,
                embedding_dims=[variable_dim] * len(d_feat_dynamic_real),
            )
        else:
            self.feat_dynamic_proj = None

        if c_feat_dynamic_cat:
            self.feat_dynamic_embed = FeatureEmbedder(
                cardinalities=c_feat_dynamic_cat,
                embedding_dims=[variable_dim] * len(c_feat_dynamic_cat),
            )
        else:
            self.feat_dynamic_embed = None

        if d_feat_static_real:
            self.feat_static_proj = FeatureProjector(
                feature_dims=d_feat_static_real,
                embedding_dims=[variable_dim] * len(d_feat_static_real),
            )
        else:
            self.feat_static_proj = None

        if c_feat_static_cat:
            self.feat_static_embed = FeatureEmbedder(
                cardinalities=c_feat_static_cat,
                embedding_dims=[variable_dim] * len(c_feat_static_cat),
            )
        else:
            self.feat_static_embed = None

        n_feat_static = len(d_feat_static_real) + len(c_feat_static_cat)
        self.static_selector = VariableSelectionNetwork(
            d_hidden=variable_dim,
            n_vars=n_feat_static,
            dropout=dropout,
        )

        n_past_feat_dynamic = len(d_past_feat_dynamic_real) + len(
            c_past_feat_dynamic_cat
        )
        n_feat_dynamic = len(d_feat_dynamic_real) + len(c_feat_dynamic_cat)
        self.ctx_selector = VariableSelectionNetwork(
            d_hidden=variable_dim,
            n_vars=n_past_feat_dynamic + n_feat_dynamic + 1,
            add_static=True,
            dropout=dropout,
        )

        self.tgt_selector = VariableSelectionNetwork(
            d_hidden=variable_dim,
            n_vars=n_feat_dynamic,
            add_static=True,
            dropout=dropout,
        )

        self.selection = GatedResidualNetwork(
            d_hidden=variable_dim,
            dropout=dropout,
        )

        self.enrichment = GatedResidualNetwork(
            d_hidden=variable_dim,
            dropout=dropout,
        )

        self.state_h = GatedResidualNetwork(
            d_hidden=variable_dim,
            d_output=embed_dim,
            dropout=dropout,
        )

        self.state_c = GatedResidualNetwork(
            d_hidden=variable_dim,
            d_output=embed_dim,
            dropout=dropout,
        )

        self.temporal_encoder = TemporalFusionEncoder(
            d_input=variable_dim,
            d_hidden=embed_dim,
        )
        self.temporal_decoder = TemporalFusionDecoder(
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            d_hidden=embed_dim,
            d_var=variable_dim,
            n_head=num_heads,
            dropout=dropout,
        )

        self.quantiles = sum(
            [[i / 10, 1.0 - i / 10] for i in range(1, (num_outputs + 1) // 2)],
            [0.5],
        )
        self.output = QuantileOutput(
            input_size=embed_dim, quantiles=self.quantiles
        )
        self.output_proj = self.output.get_quantile_proj()
        self.loss = self.output.get_loss()

    def _preprocess(
        self,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
        past_feat_dynamic_real: torch.Tensor,
        past_feat_dynamic_cat: torch.Tensor,
        feat_dynamic_real: torch.Tensor,
        feat_dynamic_cat: torch.Tensor,
        feat_static_real: torch.Tensor,
        feat_static_cat: torch.Tensor,
    ):
        obs = past_target * past_observed_values
        count = past_observed_values.sum(dim=1, keepdim=True)
        offset = obs.sum(1, keepdim=True) / (count + self.normalize_eps)
        scale = torch.sum(obs**2, 1, keepdim=True) / (
            count + self.normalize_eps
        )
        scale = torch.sqrt(scale - offset**2)

        past_target = (past_target - offset) / (scale + self.normalize_eps)
        past_target = past_target.unsqueeze(-1)

        proj = self.target_proj(past_target)

        past_covariates = []
        future_covariates = []
        static_covariates: List[torch.Tensor] = []

        past_covariates.append(proj)
        if self.past_feat_dynamic_proj is not None:
            projs = self.past_feat_dynamic_proj(past_feat_dynamic_real)
            past_covariates.extend(projs)
        if self.past_feat_dynamic_embed is not None:
            embs = self.past_feat_dynamic_embed(past_feat_dynamic_cat)
            past_covariates.extend(embs)
        if self.feat_dynamic_proj is not None:
            projs = self.feat_dynamic_proj(feat_dynamic_real)
            for proj in projs:
                ctx_proj = proj[:, 0 : self.context_length, ...]
                tgt_proj = proj[:, self.context_length :, ...]
                past_covariates.append(ctx_proj)
                future_covariates.append(tgt_proj)
        if self.feat_dynamic_embed is not None:
            embs = self.feat_dynamic_embed(feat_dynamic_cat)
            for emb in embs:
                ctx_emb = emb[:, 0 : self.context_length, ...]
                tgt_emb = emb[:, self.context_length :, ...]
                past_covariates.append(ctx_emb)
                future_covariates.append(tgt_emb)

        if self.feat_static_proj is not None:
            projs = self.feat_static_proj(feat_static_real)
            static_covariates.extend(projs)
        if self.feat_static_embed is not None:
            embs = self.feat_static_embed(feat_static_cat)
            static_covariates.extend(embs)

        return (
            past_covariates,
            future_covariates,
            static_covariates,
            offset,
            scale,
        )

    def _postprocess(
        self,
        preds: torch.Tensor,
        offset: torch.Tensor,
        scale: torch.Tensor,
    ) -> torch.Tensor:
        offset = offset.unsqueeze(-1)
        scale = scale.unsqueeze(-1)
        preds = (preds * (scale + self.normalize_eps)) + offset
        return preds

    def forward(
        self,
        past_observed_values: torch.Tensor,
        past_covariates: torch.Tensor,
        future_covariates: torch.Tensor,
        static_covariates: torch.Tensor,
    ):
        static_var, _ = self.static_selector(static_covariates)
        c_selection = self.selection(static_var).unsqueeze(1)
        c_enrichment = self.enrichment(static_var).unsqueeze(1)
        c_h = self.state_h(static_var)
        c_c = self.state_c(static_var)

        ctx_input, _ = self.ctx_selector(past_covariates, c_selection)
        tgt_input, _ = self.tgt_selector(future_covariates, c_selection)

        encoding = self.temporal_encoder(
            ctx_input, tgt_input, [c_h.unsqueeze(0), c_c.unsqueeze(0)]
        )
        decoding = self.temporal_decoder(
            encoding, c_enrichment, past_observed_values
        )

        preds = self.output_proj(decoding)

        return preds


class TemporalFusionTransformerTrainingNetwork(
    TemporalFusionTransformerNetwork
):
    def forward(
        self,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
        future_target: torch.Tensor,
        future_observed_values: torch.Tensor,
        past_feat_dynamic_real: torch.Tensor,
        past_feat_dynamic_cat: torch.Tensor,
        feat_dynamic_real: torch.Tensor,
        feat_dynamic_cat: torch.Tensor,
        feat_static_real: torch.Tensor,
        feat_static_cat: torch.Tensor,
    ) -> torch.Tensor:
        (
            past_covariates,
            future_covariates,
            static_covariates,
            offset,
            scale,
        ) = self._preprocess(
            past_target,
            past_observed_values,
            past_feat_dynamic_real,
            past_feat_dynamic_cat,
            feat_dynamic_real,
            feat_dynamic_cat,
            feat_static_real,
            feat_static_cat,
        )

        preds = super().forward(
            past_observed_values,
            past_covariates,
            future_covariates,
            static_covariates,
        )

        preds = self._postprocess(preds, offset, scale)

        loss = self.loss(future_target, preds)
        loss = weighted_average(loss, future_observed_values)
        return loss.mean()


class TemporalFusionTransformerPredictionNetwork(
    TemporalFusionTransformerNetwork
):
    def forward(
        self,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
        past_feat_dynamic_real: torch.Tensor,
        past_feat_dynamic_cat: torch.Tensor,
        feat_dynamic_real: torch.Tensor,
        feat_dynamic_cat: torch.Tensor,
        feat_static_real: torch.Tensor,
        feat_static_cat: torch.Tensor,
    ) -> torch.Tensor:
        (
            past_covariates,
            future_covariates,
            static_covariates,
            offset,
            scale,
        ) = self._preprocess(
            past_target,
            past_observed_values,
            past_feat_dynamic_real,
            past_feat_dynamic_cat,
            feat_dynamic_real,
            feat_dynamic_cat,
            feat_static_real,
            feat_static_cat,
        )

        preds = super().forward(
            past_observed_values,
            past_covariates,
            future_covariates,
            static_covariates,
        )

        preds = self._postprocess(preds, offset, scale)
        return preds.permute(0, 2, 1)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from itertools import chain
from typing import List, Optional, Dict

import numpy as np
import torch

from src.gluonts.core.component import validated
from src.gluonts.dataset.field_names import FieldName
from src.gluonts.model.forecast_generator import QuantileForecastGenerator
from src.gluonts.model.predictor import Predictor
from src.gluonts.time_feature import (
    TimeFeature,
    time_features_from_frequency_str,
)
from src.gluonts.torch.model.predictor import PyTorchPredictor
from src.gluonts.torch.util import copy_parameters
from src.gluonts.transform import (
    Transformation,
    Chain,
    ValidationSplitSampler,
    TestSplitSampler,
    ExpectedNumInstanceSampler,
    AddAgeFeature,
    AsNumpyArray,
    AddObservedValuesIndicator,
    AddTimeFeatures,
    VstackFeatures,
    SetField,
)

from pts import Trainer
from pts.model import PyTorchEstimator
from pts.model.utils import get_module_forward_input_names

from .tft_network import (
    TemporalFusionTransformerPredictionNetwork,
    TemporalFusionTransformerTrainingNetwork,
)
from .tft_transform import BroadcastTo, TFTInstanceSplitter


def _default_feat_args(dims_or_cardinalities: List[int]):
    if dims_or_cardinalities:
        return dims_or_cardinalities
    return [1]


class TemporalFusionTransformerEstimator(PyTorchEstimator):
    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        context_length: Optional[int] = None,
        dropout_rate: float = 0.1,
        embed_dim: int = 32,
        num_heads: int = 4,
        num_outputs: int = 3,
        variable_dim: Optional[int] = None,
        time_features: List[TimeFeature] = [],
        static_cardinalities: Dict[str, int] = {},
        dynamic_cardinalities: Dict[str, int] = {},
        static_feature_dims: Dict[str, int] = {},
        dynamic_feature_dims: Dict[str, int] = {},
        past_dynamic_features: List[str] = [],
        trainer: Trainer = Trainer(),
    ) -> None:
        super().__init__(trainer=trainer)

        self.freq = freq
        self.prediction_length = prediction_length
        self.context_length = context_length or prediction_length

        # MultiheadAttention
        self.dropout_rate = dropout_rate
        self.embed_dim = embed_dim
        self.num_heads = num_heads

        self.num_outputs = num_outputs
        self.variable_dim = variable_dim or embed_dim

        if not time_features:
            self.time_features = time_features_from_frequency_str(self.freq)
        else:
            self.time_features = time_features
        self.static_cardinalities = static_cardinalities
        self.dynamic_cardinalities = dynamic_cardinalities
        self.static_feature_dims = static_feature_dims
        self.dynamic_feature_dims = dynamic_feature_dims
        self.past_dynamic_features = past_dynamic_features

        self.past_dynamic_cardinalities = {}
        self.past_dynamic_feature_dims = {}
        for name in self.past_dynamic_features:
            if name in self.dynamic_cardinalities:
                self.past_dynamic_cardinalities[name] = (
                    self.dynamic_cardinalities.pop(name)
                )
            elif name in self.dynamic_feature_dims:
                self.past_dynamic_feature_dims[name] = (
                    self.dynamic_feature_dims.pop(name)
                )
            else:
                raise ValueError(
                    f"Feature name {name} is not provided in feature dicts"
                )

        self.train_sampler = ExpectedNumInstanceSampler(
            num_instances=1.0, min_future=prediction_length
        )

        self.validation_sampler = ValidationSplitSampler(
            min_future=prediction_length
        )

    def create_transformation(self) -> Transformation:
        transforms = (
            [AsNumpyArray(field=FieldName.TARGET, expected_ndim=1)]
            + (
                [
                    AsNumpyArray(field=name, expected_ndim=1)
                    for name in self.static_cardinalities.keys()
                ]
            )
            + [
                AsNumpyArray(field=name, expected_ndim=1)
                for name in chain(
                    self.static_feature_dims.keys(),
                    self.dynamic_cardinalities.keys(),
                )
            ]
            + [
                AsNumpyArray(field=name, expected_ndim=2)
                for name in self.dynamic_feature_dims.keys()
            ]
            + [
                AddObservedValuesIndicator(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.OBSERVED_VALUES,
                ),
                AddTimeFeatures(
                    start_field=FieldName.START,
                    target_field=FieldName.TARGET,
                    output_field=FieldName.FEAT_TIME,
                    time_features=self.time_features,
                    pred_length=self.prediction_length,
                ),
                AddAgeFeature(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.FEAT_AGE,
                    pred_length=self.prediction_length,
                    log_scale=True,
                ),
            ]
        )

        if self.static_cardinalities:
            transforms.extend(
                [
                    VstackFeatures(
                        output_field=FieldName.FEAT_STATIC_CAT,
                        input_fields=list(self.static_cardinalities.keys()),
                        h_stack=True,
                    ),
                    AsNumpyArray(
                        field=FieldName.FEAT_STATIC_CAT,
                        expected_ndim=1,
                        dtype=np.long,
                    ),
                ]
            )
        else:
            transforms.extend(
                [
                    SetField(
                        output_field=FieldName.FEAT_STATIC_CAT,
                        value=[0],
                    ),
                    AsNumpyArray(
                        field=FieldName.FEAT_STATIC_CAT,
                        expected_ndim=1,
                        dtype=np.long,
                    ),
                ]
            )

        if self.static_feature_dims:
            transforms.append(
                VstackFeatures(
                    output_field=FieldName.FEAT_STATIC_REAL,
                    input_fields=list(self.static_feature_dims.keys()),
                    h_stack=True,
                )
            )
        else:
            transforms.extend(
                [
                    SetField(
                        output_field=FieldName.FEAT_STATIC_REAL,
                        value=[0.0],
                    ),
                    AsNumpyArray(
                        field=FieldName.FEAT_STATIC_REAL, expected_ndim=1
                    ),
                ]
            )

        if self.dynamic_cardinalities:
            transforms.extend(
                [
                    VstackFeatures(
                        output_field=FieldName.FEAT_DYNAMIC_CAT,
                        input_fields=list(self.dynamic_cardinalities.keys()),
                    ),
                    AsNumpyArray(
                        field=FieldName.FEAT_DYNAMIC_CAT,
                        expected_ndim=2,
                        dtype=np.long,
                    ),
                ]
            )
        else:
            transforms.extend(
                [
                    SetField(
                        output_field=FieldName.FEAT_DYNAMIC_CAT,
                        value=[[0]],
                    ),
                    AsNumpyArray(
                        field=FieldName.FEAT_DYNAMIC_CAT,
                        expected_ndim=2,
                        dtype=np.long,
                    ),
                    BroadcastTo(
                        field=FieldName.FEAT_DYNAMIC_CAT,
                        ext_length=self.prediction_length,
                    ),
                ]
            )

        input_fields = [FieldName.FEAT_TIME, FieldName.FEAT_AGE]
        if self.dynamic_feature_dims:
            input_fields += list(self.dynamic_feature_dims.keys())
        transforms.append(
            VstackFeatures(
                input_fields=input_fields,
                output_field=FieldName.FEAT_DYNAMIC_REAL,
            )
        )

        if self.past_dynamic_cardinalities:
            transforms.extend(
                [
                    VstackFeatures(
                        output_field=FieldName.PAST_FEAT_DYNAMIC + "_cat",
                        input_fields=list(
                            self.past_dynamic_cardinalities.keys()
                        ),
                    ),
                    AsNumpyArray(
                        field=FieldName.PAST_FEAT_DYNAMIC + "_cat",
                        expected_ndim=2,
                        dtype=np.long,
                    ),
                ]
            )
        else:
            transforms.extend(
                [
                    SetField(
                        output_field=FieldName.PAST_FEAT_DYNAMIC + "_cat",
                        value=[[0]],
                    ),
                    AsNumpyArray(
                        field=FieldName.PAST_FEAT_DYNAMIC + "_cat",
                        expected_ndim=2,
                        dtype=np.long,
                    ),
                    BroadcastTo(field=FieldName.PAST_FEAT_DYNAMIC + "_cat"),
                ]
            )

        if self.past_dynamic_feature_dims:
            transforms.append(
                VstackFeatures(
                    output_field=FieldName.PAST_FEAT_DYNAMIC_REAL,
                    input_fields=list(self.past_dynamic_feature_dims.keys()),
                )
            )
        else:
            transforms.extend(
                [
                    SetField(
                        output_field=FieldName.PAST_FEAT_DYNAMIC_REAL,
                        value=[[0.0]],
                    ),
                    AsNumpyArray(
                        field=FieldName.PAST_FEAT_DYNAMIC_REAL, expected_ndim=2
                    ),
                    BroadcastTo(field=FieldName.PAST_FEAT_DYNAMIC_REAL),
                ]
            )

        return Chain(transforms)

    def create_instance_splitter(self, mode: str):
        assert mode in ["training", "validation", "test"]

        instance_sampler = {
            "training": self.train_sampler,
            "validation": self.validation_sampler,
            "test": TestSplitSampler(),
        }[mode]

        ts_fields = [FieldName.FEAT_DYNAMIC_CAT, FieldName.FEAT_DYNAMIC_REAL]
        past_ts_fields = [
            FieldName.PAST_FEAT_DYNAMIC + "_cat",
            FieldName.PAST_FEAT_DYNAMIC_REAL,
        ]

        return TFTInstanceSplitter(
            instance_sampler=instance_sampler,
            past_length=self.context_length,
            future_length=self.prediction_length,
            time_series_fields=ts_fields,
            past_time_series_fields=past_ts_fields,
        )

    def create_training_network(
        self, device: torch.device
    ) -> TemporalFusionTransformerTrainingNetwork:
        network = TemporalFusionTransformerTrainingNetwork(
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            variable_dim=self.variable_dim,
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_outputs=self.num_outputs,
            dropout=self.dropout_rate,
            d_past_feat_dynamic_real=_default_feat_args(
                list(self.past_dynamic_feature_dims.values())
            ),
            c_past_feat_dynamic_cat=_default_feat_args(
                list(self.past_dynamic_cardinalities.values())
            ),
            # +1 is for Age Feature
            d_feat_dynamic_real=_default_feat_args(
                [1] * (len(self.time_features) + 1)
                + list(self.dynamic_feature_dims.values())
            ),
            c_feat_dynamic_cat=_default_feat_args(
                list(self.dynamic_cardinalities.values())
            ),
            d_feat_static_real=_default_feat_args(
                list(self.static_feature_dims.values()),
            ),
            c_feat_static_cat=_default_feat_args(
                list(self.static_cardinalities.values()),
            ),
        )
        return network.to(device)

    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: TemporalFusionTransformerTrainingNetwork,
        device: torch.device,
    ) -> Predictor:
        prediction_network = TemporalFusionTransformerPredictionNetwork(
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            variable_dim=self.variable_dim,
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_outputs=self.num_outputs,
            dropout=self.dropout_rate,
            d_past_feat_dynamic_real=_default_feat_args(
                list(self.past_dynamic_feature_dims.values())
            ),
            c_past_feat_dynamic_cat=_default_feat_args(
                list(self.past_dynamic_cardinalities.values())
            ),
            # +1 is for Age Feature
            d_feat_dynamic_real=_default_feat_args(
                [1] * (len(self.time_features) + 1)
                + list(self.dynamic_feature_dims.values())
            ),
            c_feat_dynamic_cat=_default_feat_args(
                list(self.dynamic_cardinalities.values())
            ),
            d_feat_static_real=_default_feat_args(
                list(self.static_feature_dims.values()),
            ),
            c_feat_static_cat=_default_feat_args(
                list(self.static_cardinalities.values()),
            ),
        ).to(device)

        copy_parameters(trained_network, prediction_network)
        input_names = get_module_forward_input_names(prediction_network)
        prediction_splitter = self.create_instance_splitter("test")

        return PyTorchPredictor(
            input_transform=transformation + prediction_splitter,
            input_names=input_names,
            prediction_net=prediction_network,
            batch_size=self.trainer.batch_size,
            freq=self.freq,
            prediction_length=self.prediction_length,
            device=device,
            forecast_generator=QuantileForecastGenerator(
                quantiles=[str(q) for q in prediction_network.quantiles],
            ),
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import time
from typing import List, Optional, Union

from tqdm.auto import tqdm

import torch
import torch.nn as nn
from torch.optim import Adam
from torch.optim.lr_scheduler import OneCycleLR
from torch.utils.data import DataLoader

from src.gluonts.core.component import validated


def requires_grad_(model: torch.nn.Module, requires_grad: bool) -> None:
    for param in model.parameters():
        param.requires_grad_(requires_grad)


def change_device(var, device):
    if isinstance(var, torch.Tensor):
        if var.device == "cpu":
            var.to(device)
            return var
        elif var.device != device:
            return var.cpu().to(device)
        else:
            return var
    if isinstance(var, dict):
        for key in var.keys():
            if isinstance(var[key], torch.Tensor):
                var[key] = var[key].to(device)
        return var

    return torch.from_numpy(var).float().to(device)


class Trainer:
    @validated()
    def __init__(
        self,
        epochs: int = 100,
        batch_size: int = 32,
        num_batches_per_epoch: int = 50,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-6,
        maximum_learning_rate: float = 1e-2,
        clip_gradient: Optional[float] = None,
        device: Optional[Union[torch.device, str]] = None,
        **kwargs,
    ) -> None:
        self.epochs = epochs
        self.batch_size = batch_size
        self.num_batches_per_epoch = num_batches_per_epoch
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.maximum_learning_rate = maximum_learning_rate
        self.clip_gradient = clip_gradient
        self.device = device

    def __call__(
        self,
        net: nn.Module,
        train_iter: DataLoader,
        validation_iter: Optional[DataLoader] = None,
    ) -> None:
        optimizer = Adam(
            net.parameters(),
            lr=self.learning_rate,
            weight_decay=self.weight_decay,
        )

        lr_scheduler = OneCycleLR(
            optimizer,
            max_lr=self.maximum_learning_rate,
            steps_per_epoch=self.num_batches_per_epoch,
            epochs=self.epochs,
        )

        for epoch_no in range(self.epochs):
            # mark epoch start time
            tic = time.time()
            cumm_epoch_loss = 0.0
            total = self.num_batches_per_epoch - 1

            # training loop
            with tqdm(train_iter, total=total) as it:
                for batch_no, data_entry in enumerate(it, start=1):
                    optimizer.zero_grad()
                    inputs = []
                    for v in data_entry.values():
                        if v.ndim >= 4:
                            v = v.reshape(v.shape[:3])
                        inputs.append(v.to(self.device))
                    # inputs = [v.to(self.device) for v in data_entry.values()]

                    output = net(*inputs)

                    if isinstance(output, (list, tuple)):
                        loss = output[0]
                    else:
                        loss = output

                    cumm_epoch_loss += loss.item()
                    avg_epoch_loss = cumm_epoch_loss / batch_no
                    it.set_postfix(
                        {
                            "epoch": f"{epoch_no + 1}/{self.epochs}",
                            "avg_loss": avg_epoch_loss,
                        },
                        refresh=False,
                    )

                    loss.backward()
                    if self.clip_gradient is not None:
                        nn.utils.clip_grad_norm_(
                            net.parameters(), self.clip_gradient
                        )

                    optimizer.step()
                    lr_scheduler.step()

                    if self.num_batches_per_epoch == batch_no:
                        break
                it.close()

            # validation loop
            if validation_iter is not None:
                cumm_epoch_loss_val = 0.0
                with tqdm(validation_iter, total=total, colour="green") as it:
                    for batch_no, data_entry in enumerate(it, start=1):
                        inputs = [
                            v.to(self.device) for v in data_entry.values()
                        ]
                        with torch.no_grad():
                            output = net(*inputs)
                        if isinstance(output, (list, tuple)):
                            loss = output[0]
                        else:
                            loss = output

                        cumm_epoch_loss_val += loss.item()
                        avg_epoch_loss_val = cumm_epoch_loss_val / batch_no
                        it.set_postfix(
                            {
                                "epoch": f"{epoch_no + 1}/{self.epochs}",
                                "avg_loss": avg_epoch_loss,
                                "avg_val_loss": avg_epoch_loss_val,
                            },
                            refresh=False,
                        )

                        if self.num_batches_per_epoch == batch_no:
                            break

                it.close()

            # mark epoch end time and log time cost of current epoch
            toc = time.time()


class Trainer_adv(Trainer):
    @validated()
    def __init__(
        self,
        sparse_net,
        noise_sd=0.1,
        clamp=False,
        epochs: int = 100,
        batch_size: int = 32,
        num_batches_per_epoch: int = 50,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-6,
        maximum_learning_rate: float = 1e-2,
        clip_gradient: Optional[float] = None,
        device: Optional[Union[torch.device, str]] = None,
        **kwargs,
    ) -> None:
        self.epochs = epochs
        self.batch_size = batch_size
        self.num_batches_per_epoch = num_batches_per_epoch
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.maximum_learning_rate = maximum_learning_rate
        self.clip_gradient = clip_gradient
        self.device = device
        self.sparse_net = sparse_net
        self.noise_sd = noise_sd
        self.clamp = clamp

    def __call__(
        self,
        net: nn.Module,
        train_iter: DataLoader,
        validation_iter: Optional[DataLoader] = None,
    ) -> None:
        optimizer = Adam(
            net.parameters(),
            lr=self.learning_rate,
            weight_decay=self.weight_decay,
        )
        optimizer_adv = Adam(
            self.sparse_net.parameters(),
            lr=self.learning_rate,
            weight_decay=self.weight_decay,
        )

        lr_scheduler = OneCycleLR(
            optimizer,
            max_lr=self.maximum_learning_rate,
            steps_per_epoch=self.num_batches_per_epoch,
            epochs=self.epochs,
        )
        loss_function = nn.MSELoss()

        for epoch_no in range(self.epochs):
            # mark epoch start time
            tic = time.time()
            cumm_epoch_loss = 0.0
            total = self.num_batches_per_epoch - 1

            # training loop
            with tqdm(train_iter, total=total) as it:
                for batch_no, data_entry in enumerate(it, start=1):
                    # train sparse net
                    self.sparse_net.zero_grad()
                    inputs = {
                        key: data_entry[key].to(self.device)
                        for key in data_entry.keys()
                    }
                    shapes = data_entry["past_target_cdf"].shape
                    delta = self.sparse_net(
                        data_entry["past_target_cdf"].to(self.device),
                        n_sample=100,
                    ).view(shapes)
                    if self.clamp:
                        delta = torch.clamp(
                            delta,
                            max=self.sparse_net.max_norm
                            * data_entry["past_target_cdf"].abs().max(),
                        )
                    inputs["past_target_cdf"] += delta
                    output = net(**inputs)
                    mu, target = output[2], output[-1]
                    loss_sparse = -loss_function(mu, target)
                    loss_sparse.backward()
                    optimizer_adv.step()

                    # train forecasting model
                    net.zero_grad()
                    perturbed_inputs = {
                        key: data_entry[key].to(self.device)
                        for key in data_entry.keys()
                    }
                    shapes = data_entry["past_target_cdf"].shape
                    delta = self.sparse_net(
                        data_entry["past_target_cdf"].to(self.device),
                        n_sample=100,
                    ).view(shapes)
                    perturbed_inputs["past_target_cdf"] += delta
                    output = net(**perturbed_inputs)

                    if isinstance(output, (list, tuple)):
                        loss = output[0]
                    else:
                        loss = output
                    cumm_epoch_loss += loss.item()
                    avg_epoch_loss = cumm_epoch_loss / batch_no
                    it.set_postfix(
                        {
                            "epoch": f"{epoch_no + 1}/{self.epochs}",
                            "avg_loss": avg_epoch_loss,
                        },
                        refresh=False,
                    )

                    loss.backward()
                    if self.clip_gradient is not None:
                        nn.utils.clip_grad_norm_(
                            net.parameters(), self.clip_gradient
                        )

                    optimizer.step()
                    lr_scheduler.step()

                    if self.num_batches_per_epoch == batch_no:
                        break

                it.close()

            # validation loop
            if validation_iter is not None:
                cumm_epoch_loss_val = 0.0
                with tqdm(validation_iter, total=total, colour="green") as it:
                    for batch_no, data_entry in enumerate(it, start=1):
                        inputs = [
                            v.to(self.device) for v in data_entry.values()
                        ]
                        with torch.no_grad():
                            output = net(*inputs)
                        if isinstance(output, (list, tuple)):
                            loss = output[0]
                        else:
                            loss = output

                        cumm_epoch_loss_val += loss.item()
                        avg_epoch_loss_val = cumm_epoch_loss_val / batch_no
                        it.set_postfix(
                            {
                                "epoch": f"{epoch_no + 1}/{self.epochs}",
                                "avg_loss": avg_epoch_loss,
                                "avg_val_loss": avg_epoch_loss_val,
                            },
                            refresh=False,
                        )

                        if self.num_batches_per_epoch == batch_no:
                            break

                it.close()

            # mark epoch end time and log time cost of current epoch
            toc = time.time()
            # print('Epoch: {}\tTime: {:.1f}'.format(epoch_no, toc - tic))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import NamedTuple, Callable, Optional, Iterator
import re

from torch import Tensor
from torch._six import container_abcs, int_classes, string_classes
from torch.utils.data import Dataset, DataLoader, DistributedSampler
import torch as pt

from ..engine.distributed import is_distributed, get_world_size

_default_collate_err_msg_format = (
    "_default_collate: batch must contain tensors, numpy arrays, numbers, "
    "dicts or lists; found {}"
)
np_str_obj_array_pattern = re.compile(r"[SaUO]")


def _default_collate(batch):
    r"""
    Puts each data field into a tensor with outer dimension batch size.
    """

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, pt.Tensor):
        out = None
        if pt.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum([x.numel() for x in batch])
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return pt.stack(batch, 0, out=out)
    elif (
        elem_type.__module__ == "numpy"
        and elem_type.__name__ != "str_"
        and elem_type.__name__ != "string_"
    ):
        elem = batch[0]
        if elem_type.__name__ == "ndarray":
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(
                    _default_collate_err_msg_format.format(elem.dtype)
                )
            return _default_collate([pt.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return pt.as_tensor(batch)
    elif isinstance(elem, float):
        return pt.tensor(batch, dtype=pt.float)
    elif isinstance(elem, int_classes):
        return pt.tensor(batch, dtype=pt.long)
    elif isinstance(elem, string_classes):
        return batch
    elif isinstance(elem, container_abcs.Mapping):
        return {key: _default_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, "_fields"):  # namedtuple
        return elem_type(
            *(_default_collate(samples) for samples in zip(*batch))
        )
    elif isinstance(elem, container_abcs.Sequence):
        transposed = zip(*batch)
        return [_default_collate(samples) for samples in transposed]
    elif elem is None:
        return None

    raise TypeError(_default_collate_err_msg_format.format(elem_type))


def copy_to_gpu(data, cuda_device: int, non_blocking: bool):
    if not pt.cuda.is_available():
        raise SystemError("GPU is not available!")
    if isinstance(data, pt.Tensor):
        return data.cuda(cuda_device, non_blocking=non_blocking)
    elif isinstance(data, container_abcs.Mapping):
        return {
            key: copy_to_gpu(val, cuda_device, non_blocking)
            for key, val in data.items()
        }
    elif isinstance(data, container_abcs.Sequence):
        return [copy_to_gpu(t, cuda_device, non_blocking) for t in data]
    elif data is None:
        return None
    else:
        raise ValueError(
            f"expected tensor, sequence or dictionary of tensors, received {type(data).__name__}"
        )


class MetaDataset(NamedTuple):
    """
    Dataset Split Manager. Possess train/valid/test datasets and provide data
    loaders.

    Parameters:
    --------------
    train_set: torch.utils.data.Dataset
        training dataset
    valid_set: torch.utils.data.Dataset or None
        validation dataset if provided
    test_set: torch.utils.data.Dataset
        test dataset
    collate_fn: callable
        function that accepts a batch of inputs and
        returns a stacked version for each data field
    always_validation: bool
        if True, use test data at request of validation loader;
        otherwise return None
    """

    train_set: Dataset
    valid_set: Optional[Dataset]
    test_set: Dataset
    collate_fn: Callable[..., Optional[Tensor]] = _default_collate
    always_validation: bool = False

    def __getattr__(self, key: str):
        return getattr(self.train_set, key)

    @property
    def train_size(self) -> int:
        return len(self.train_set)

    @property
    def valid_size(self) -> Optional[int]:
        if self.valid_set is None:
            raise AttributeError("No validation set")
        return len(self.valid_set)

    @property
    def test_size(self) -> int:
        return len(self.test_set)

    def _data_loader(
        self,
        dataset: Dataset,
        batch_size: int,
        shuffle: bool,
        cuda_device: int,
        is_training: bool,
        n_workers: int,
        n_batches: Optional[int],
    ) -> Iterator:
        if is_distributed() and is_training:
            sampler = DistributedSampler(dataset, shuffle=shuffle)
            batch_size //= get_world_size()
            shuffle = False
        else:
            sampler = None
        loader = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            sampler=sampler,
            num_workers=n_workers,
            collate_fn=self.collate_fn,
            pin_memory=cuda_device >= 0,
            drop_last=n_batches is not None,
        )
        if n_batches is None:
            n_batches = len(loader)
        iterator = iter(loader)
        for batch in range(n_batches):
            try:
                data = next(iterator)
            except StopIteration:
                iterator = iter(loader)
                data = next(iterator)
            if cuda_device >= 0:
                data = copy_to_gpu(data, cuda_device, non_blocking=True)
            yield data

    def train_loader(
        self,
        batch_size: int,
        shuffle: bool,
        cuda_device: int,
        n_workers: int = 0,
        n_batches: Optional[int] = None,
    ) -> Iterator:
        return self._data_loader(
            self.train_set,
            batch_size,
            shuffle,
            cuda_device,
            True,
            n_workers,
            n_batches,
        )

    def valid_loader(
        self,
        batch_size: int,
        cuda_device: int,
        n_workers: int = 0,
    ) -> Iterator:
        if self.valid_set is None:
            if self.always_validation:
                dataset = self.test_set
            else:
                raise AttributeError("No validation set")
        else:
            dataset = self.valid_set
        return self._data_loader(
            dataset,
            batch_size,
            False,
            cuda_device,
            False,
            n_workers,
            None,
        )

    def test_loader(
        self,
        batch_size: int,
        cuda_device: int,
        n_workers: int = 0,
    ) -> Iterator:
        return self._data_loader(
            self.test_set,
            batch_size,
            False,
            cuda_device,
            False,
            n_workers,
            None,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import List, Optional
from collections import deque
import math

import torch as pt
from torch import Tensor
from torch import nn
from torch.distributions import Distribution
from torch.distributions import Normal as Gaussian


class ResidualBlock(nn.Module):
    """
    Network module wrapped by residual connection.

    Args
    ----------
    residual_network: nn.Module
        the main module to be wrapped
    skip_network: nn.Module, optional
        the auxiliary module to guarantee valid addition
    """

    def __init__(
        self,
        residual_network: nn.Module,
        skip_network: Optional[nn.Module] = None,
    ):
        super(ResidualBlock, self).__init__()
        self.residual_network = residual_network
        if skip_network is None:
            self.skip_network = nn.Identity()
        else:
            self.skip_network = skip_network

    def forward(self, x: Tensor, *args, **kwargs) -> Tensor:
        skip = self.skip_network(x)
        x = self.residual_network(x, *args, **kwargs)
        return skip + x


class Concatenate(nn.Module):
    def __init__(self, dim: int):
        super(Concatenate, self).__init__()
        self.dim = dim

    def forward(self, inputs: List[Tensor]) -> Tensor:
        return pt.cat(inputs, dim=self.dim)


class LockedDropout(nn.Module):
    def __init__(self, dropout=0, lock_dim=None, inplace=False):
        super(LockedDropout, self).__init__()
        self.dropout = nn.Dropout(p=dropout, inplace=inplace)
        self.lock_dim = lock_dim

    def forward(self, x):
        if self.lock_dim is None:
            return self.dropout(x)
        else:
            mask = self.dropout(
                pt.ones_like(x).narrow(self.lock_dim, 0, 1)
            ).expand_as(x)
            return mask * x


class DilatedQueue(nn.Module):
    def __init__(self, maxlen):
        super(DilatedQueue, self).__init__()
        self.maxlen = maxlen
        self.register_buffer("data", None)

    def _reset(self):
        self.data = None

    def initialize(self, x):
        self._reset()
        if x.size(2) < self.maxlen:
            x = nn.functional.pad(x, (self.maxlen - x.size(2), 0))
        else:
            x = x.narrow(2, x.size(2) - self.maxlen, self.maxlen)
        self.data = x

    def append(self, x):
        data = pt.cat([self.data, x], dim=2)
        self.data = data.narrow(2, data.size(2) - self.maxlen, self.maxlen)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, Type, TypeVar, Any
import pytorch_lightning as pl
from src.gluonts.dataset.repository.datasets import dataset_recipes


DATA_MODULE_REGISTRY: Dict[str, Type[pl.LightningDataModule]] = {}


M = TypeVar("M", bound=Type[pl.LightningDataModule])


def register_data_module(cls: M) -> M:
    """
    Registers the provided class in the global model registry.
    """
    DATA_MODULE_REGISTRY[cls.name()] = cls
    return cls


def get_data_module(name: str, **kwargs: Any) -> pl.LightningDataModule:
    """
    This method creates the data module with the specified name. The provided
    keyword arguments must contain ALL arguments required by the model
    configuration. Superfluous arguments may be provided and are simply
    ignored.

    Parameters
    ----------
    name: str
        The canonical name of the data module. See `DATA_MODULE_REGISTRY`.
    kwargs: Any
        Keyword arguments passed to the initializer of the data module.

    Returns
    -------
    ModelConfig
        The data module.
    """
    if name in DATA_MODULE_REGISTRY:
        dm_module_cls = DATA_MODULE_REGISTRY[name]
    elif name.startswith("dm_m4"):
        dm_module_cls = DATA_MODULE_REGISTRY["dm_m4"]
    elif name.startswith("dm_m3"):
        dm_module_cls = DATA_MODULE_REGISTRY["dm_m3"]
    elif name.startswith("dm_m1"):
        dm_module_cls = DATA_MODULE_REGISTRY["dm_m1"]
    elif name[3:] in dataset_recipes.keys():
        dm_module_cls = DATA_MODULE_REGISTRY["dm_gluonts"]
    else:
        print(f"Data module name {name} is unknown.")

    return dm_module_cls(**kwargs)
    # return dm_module_cls(
    #     **{k: v for k, v in kwargs.items() if k in dm_module_cls.__init__.__annotations__}
    # )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import math

import torch
from torch import nn
import torch.nn.functional as F


class DiffusionEmbedding(nn.Module):
    def __init__(self, dim, proj_dim, max_steps=500):
        super().__init__()
        self.register_buffer(
            "embedding",
            self._build_embedding(dim, max_steps),
            persistent=False,
        )
        self.projection1 = nn.Linear(dim * 2, proj_dim)
        self.projection2 = nn.Linear(proj_dim, proj_dim)

    def forward(self, diffusion_step):
        x = self.embedding[diffusion_step]
        x = self.projection1(x)
        x = F.silu(x)
        x = self.projection2(x)
        x = F.silu(x)
        return x

    def _build_embedding(self, dim, max_steps):
        steps = torch.arange(max_steps).unsqueeze(1)  # [T,1]
        dims = torch.arange(dim).unsqueeze(0)  # [1,dim]
        table = steps * 10.0 ** (dims * 4.0 / dim)  # [T,dim]
        table = torch.cat([torch.sin(table), torch.cos(table)], dim=1)
        return table


class ResidualBlock(nn.Module):
    def __init__(self, hidden_size, residual_channels, dilation):
        super().__init__()
        self.dilated_conv = nn.Conv1d(
            residual_channels,
            2 * residual_channels,
            3,
            padding=dilation,
            dilation=dilation,
            padding_mode="circular",
        )
        self.diffusion_projection = nn.Linear(hidden_size, residual_channels)
        self.conditioner_projection = nn.Conv1d(
            1, 2 * residual_channels, 1, padding=2, padding_mode="circular"
        )
        self.output_projection = nn.Conv1d(
            residual_channels, 2 * residual_channels, 1
        )

        nn.init.kaiming_normal_(self.conditioner_projection.weight)
        nn.init.kaiming_normal_(self.output_projection.weight)

    def forward(self, x, conditioner, diffusion_step):
        diffusion_step = self.diffusion_projection(diffusion_step).unsqueeze(
            -1
        )
        conditioner = self.conditioner_projection(conditioner)

        y = x + diffusion_step
        y = self.dilated_conv(y) + conditioner

        gate, filter = torch.chunk(y, 2, dim=1)
        y = torch.sigmoid(gate) * torch.tanh(filter)

        y = self.output_projection(y)
        y = F.leaky_relu(y, 0.4)
        residual, skip = torch.chunk(y, 2, dim=1)
        return (x + residual) / math.sqrt(2.0), skip


class CondUpsampler(nn.Module):
    def __init__(self, cond_length, target_dim):
        super().__init__()
        self.linear1 = nn.Linear(cond_length, target_dim // 2)
        self.linear2 = nn.Linear(target_dim // 2, target_dim)

    def forward(self, x):
        x = self.linear1(x)
        x = F.leaky_relu(x, 0.4)
        x = self.linear2(x)
        x = F.leaky_relu(x, 0.4)
        return x


class EpsilonTheta(nn.Module):
    def __init__(
        self,
        target_dim,
        cond_length,
        time_emb_dim=16,
        residual_layers=8,
        residual_channels=8,
        dilation_cycle_length=2,
        residual_hidden=64,
    ):
        super().__init__()
        self.input_projection = nn.Conv1d(
            1, residual_channels, 1, padding=2, padding_mode="circular"
        )
        self.diffusion_embedding = DiffusionEmbedding(
            time_emb_dim, proj_dim=residual_hidden
        )
        self.cond_upsampler = CondUpsampler(
            target_dim=target_dim, cond_length=cond_length
        )
        self.residual_layers = nn.ModuleList(
            [
                ResidualBlock(
                    residual_channels=residual_channels,
                    dilation=2 ** (i % dilation_cycle_length),
                    hidden_size=residual_hidden,
                )
                for i in range(residual_layers)
            ]
        )
        self.skip_projection = nn.Conv1d(
            residual_channels, residual_channels, 3
        )
        self.output_projection = nn.Conv1d(residual_channels, 1, 3)

        nn.init.kaiming_normal_(self.input_projection.weight)
        nn.init.kaiming_normal_(self.skip_projection.weight)
        nn.init.zeros_(self.output_projection.weight)

    def forward(self, inputs, time, cond):
        x = self.input_projection(inputs)
        x = F.leaky_relu(x, 0.4)

        diffusion_step = self.diffusion_embedding(time)
        cond_up = self.cond_upsampler(cond)
        skip = []
        for layer in self.residual_layers:
            x, skip_connection = layer(x, cond_up, diffusion_step)
            skip.append(skip_connection)

        x = torch.sum(torch.stack(skip), dim=0) / math.sqrt(
            len(self.residual_layers)
        )
        x = self.skip_projection(x)
        x = F.leaky_relu(x, 0.4)
        x = self.output_projection(x)
        return x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
src.gluonts.nursery.glide
~~~~~~~~~~~~~~~~~~~~~

Glide is a simple pipeline, which is able to map functions in parallel.

The core idea is to work on already sharded input-data, since inter-process
communication is expensive using multiprocessing. Thus, glide's only
syncronisation point is a final result queue to which each process emits its
data.

To help splitting the input preemptively glides offers a `partition`-method.

    from src.gluonts.nursery import glide

    data = range(100)
    parts = glide.partition(data, 3)

    def double(n):
        return n * 2

    assert set(glide.ParApply(glide.lift(double), parts)) == set(range(0, 200, 2))

"""

__all__ = [
    "Apply",
    "divide_into",
    "imap_unordered",
    "lift",
    "Map",
    "ParApply",
    "partition",
    "Pipeline",
]

from ._partition import partition, divide_into
from .parallel import ParApply
from .sequential import Apply
from .pipeline import Pipeline
from .util import lift, Map


def imap_unordered(fn, data, num_workers, batch_size=1):
    return ParApply(
        lift(fn), partition(data, num_workers), batch_size=batch_size
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import NamedTuple, Optional
from functools import partial

import numpy as np

import torch
import torch.nn as nn
from torch.utils import data
from torch.utils.data import DataLoader

from src.gluonts.env import env
from src.gluonts.core.component import validated
from src.gluonts.dataset.common import Dataset
from src.gluonts.model.estimator import Estimator
from src.gluonts.torch.model.predictor import PyTorchPredictor
from src.gluonts.transform import SelectFields, Transformation

# from src.gluonts.support.util import maybe_len

from pts import Trainer
from pts.model import get_module_forward_input_names
from pts.dataset.loader import TransformedIterableDataset


def maybe_len(obj) -> Optional[int]:
    try:
        return len(obj)
    except (NotImplementedError, AttributeError):
        return None


class TrainOutput(NamedTuple):
    transformation: Transformation
    trained_net: nn.Module
    predictor: PyTorchPredictor


class PyTorchEstimator(Estimator):
    @validated()
    def __init__(
        self,
        trainer: Trainer,
        lead_time: int = 0,
        dtype: np.dtype = np.float32,
    ) -> None:
        super().__init__(lead_time=lead_time)
        self.trainer = trainer
        self.dtype = dtype

    def create_transformation(self) -> Transformation:
        """
        Create and return the transformation needed for training and inference.

        Returns
        -------
        Transformation
            The transformation that will be applied entry-wise to datasets,
            at training and inference time.
        """
        raise NotImplementedError

    def create_instance_splitter(self, mode: str) -> Transformation:
        """
        Create and return the instance splitter needed for training, validation
        or testing.

        Returns
        -------
        Transformation
            The InstanceSplitter that will be applied entry-wise to datasets,
            at training, validation and inference time based on mode.
        """
        raise NotImplementedError

    def create_training_network(self, device: torch.device) -> nn.Module:
        """
        Create and return the network used for training (i.e., computing the
        loss).

        Returns
        -------
        nn.Module
            The network that computes the loss given input data.
        """
        raise NotImplementedError

    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: nn.Module,
        device: torch.device,
    ) -> PyTorchPredictor:
        """
        Create and return a predictor object.

        Returns
        -------
        Predictor
            A predictor wrapping a `nn.Module` used for inference.
        """
        raise NotImplementedError

    def get_loader(
        self,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
        num_workers: int = 0,
        prefetch_factor: int = 2,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
        **kwargs,
    ) -> TrainOutput:
        transformation = self.create_transformation()

        trained_net = self.create_training_network(self.trainer.device)

        input_names = get_module_forward_input_names(trained_net)

        with env._let(max_idle_transforms=maybe_len(training_data) or 0):
            training_instance_splitter = self.create_instance_splitter(
                "training"
            )
        training_iter_dataset = TransformedIterableDataset(
            dataset=training_data,
            transform=transformation
            + training_instance_splitter
            + SelectFields(input_names),
            is_train=True,
            shuffle_buffer_length=shuffle_buffer_length,
            cache_data=cache_data,
        )

        training_data_loader = DataLoader(
            training_iter_dataset,
            batch_size=self.trainer.batch_size,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            pin_memory=True,
            worker_init_fn=self._worker_init_fn,
            **kwargs,
        )
        return training_data_loader

    def train_model(
        self,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
        num_workers: int = 0,
        prefetch_factor: int = 2,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
        **kwargs,
    ) -> TrainOutput:
        transformation = self.create_transformation()

        trained_net = self.create_training_network(self.trainer.device)

        input_names = get_module_forward_input_names(trained_net)

        with env._let(max_idle_transforms=maybe_len(training_data) or 0):
            training_instance_splitter = self.create_instance_splitter(
                "training"
            )
        training_iter_dataset = TransformedIterableDataset(
            dataset=training_data,
            transform=transformation
            + training_instance_splitter
            + SelectFields(input_names),
            is_train=True,
            shuffle_buffer_length=shuffle_buffer_length,
            cache_data=cache_data,
        )

        training_data_loader = DataLoader(
            training_iter_dataset,
            batch_size=self.trainer.batch_size,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            pin_memory=True,
            worker_init_fn=self._worker_init_fn,
            **kwargs,
        )

        validation_data_loader = None
        if validation_data is not None:
            with env._let(max_idle_transforms=maybe_len(validation_data) or 0):
                validation_instance_splitter = self.create_instance_splitter(
                    "validation"
                )
            validation_iter_dataset = TransformedIterableDataset(
                dataset=validation_data,
                transform=transformation
                + validation_instance_splitter
                + SelectFields(input_names),
                is_train=True,
                cache_data=cache_data,
            )
            validation_data_loader = DataLoader(
                validation_iter_dataset,
                batch_size=self.trainer.batch_size,
                num_workers=num_workers,
                prefetch_factor=prefetch_factor,
                pin_memory=True,
                worker_init_fn=self._worker_init_fn,
                **kwargs,
            )

        self.trainer(
            net=trained_net,
            train_iter=training_data_loader,
            validation_iter=validation_data_loader,
        )

        return TrainOutput(
            transformation=transformation,
            trained_net=trained_net,
            predictor=self.create_predictor(
                transformation, trained_net, self.trainer.device
            ),
        )

    @staticmethod
    def _worker_init_fn(worker_id):
        np.random.seed(np.random.get_state()[1][0] + worker_id)

    def train(
        self,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
        num_workers: int = 0,
        prefetch_factor: int = 2,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
        **kwargs,
    ) -> PyTorchPredictor:
        return self.train_model(
            training_data,
            validation_data,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            shuffle_buffer_length=shuffle_buffer_length,
            cache_data=cache_data,
            **kwargs,
        ).predictor



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial
from pathlib import Path
from typing import Optional
import click
from tsbench.analysis.utils import num_fitting_processes, run_parallel
from tsbench.config import DATASET_REGISTRY
from tsbench.constants import DEFAULT_DATA_PATH
from ._main import datasets


@datasets.command(short_help="Download and preprocess datasets.")
@click.option(
    "--dataset",
    type=str,
    default=None,
    help=(
        "The dataset to compute basic statistics for. "
        "If not provided, computes statistics for all datasets."
    ),
)
@click.option(
    "--path",
    type=str,
    default=DEFAULT_DATA_PATH,
    show_default=True,
    help="The path where the datasets should be downloaded to.",
)
def download(dataset: Optional[str], path: str):
    """
    Downloads and preprocesses either a single dataset or all datasets in the
    registry.
    """
    base = Path(path)

    if dataset is not None:
        dataset_cls = DATASET_REGISTRY[dataset](base)
        dataset_cls.generate()
        dataset_cls.prepare()
        return

    # Start off by downloading an M3 dataset
    dataset_cls = DATASET_REGISTRY["m3_monthly"](base)
    dataset_cls.generate()
    dataset_cls.prepare()

    # Then, we can download the rest in parallel (by preloading, we don't download the M3 data in
    # parallel)
    run_parallel(
        partial(_download_dataset, base=base),
        list(DATASET_REGISTRY.keys()),
        num_processes=min(
            num_fitting_processes(cpus_per_process=1, memory_per_process=8),
            len(DATASET_REGISTRY),
        ),
    )


def _download_dataset(name: str, base: Path):
    dataset_cls = DATASET_REGISTRY[name](base)
    dataset_cls.generate()
    dataset_cls.prepare()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List

import torch
import torch.nn as nn
from torch.distributions import Distribution

from src.gluonts.core.component import validated
from src.gluonts.torch.modules.distribution_output import DistributionOutput
from src.gluonts.torch.modules.lambda_layer import LambdaLayer
from pts.modules import MeanScaler, NOPScaler


class SimpleFeedForwardNetworkBase(nn.Module):
    """
    Abstract base class to implement feed-forward networks for probabilistic
    time series prediction.

    This class does not implement hybrid_forward: this is delegated
    to the two subclasses SimpleFeedForwardTrainingNetwork and
    SimpleFeedForwardPredictionNetwork, that define respectively how to
    compute the loss and how to generate predictions.

    Parameters
    ----------
    num_hidden_dimensions
        Number of hidden nodes in each layer.
    prediction_length
        Number of time units to predict.
    context_length
        Number of time units that condition the predictions.
    batch_normalization
        Whether to use batch normalization.
    mean_scaling
        Scale the network input by the data mean and the network output by
        its inverse.
    distr_output
        Distribution to fit.
    kwargs
    """

    @validated()
    def __init__(
        self,
        num_hidden_dimensions: List[int],
        prediction_length: int,
        context_length: int,
        batch_normalization: bool,
        mean_scaling: bool,
        distr_output: DistributionOutput,
    ) -> None:
        super().__init__()

        self.num_hidden_dimensions = num_hidden_dimensions
        self.prediction_length = prediction_length
        self.context_length = context_length
        self.batch_normalization = batch_normalization
        self.mean_scaling = mean_scaling
        self.distr_output = distr_output

        modules = []
        dims = self.num_hidden_dimensions
        for i, units in enumerate(dims[:-1]):
            if i == 0:
                input_size = context_length
            else:
                input_size = dims[i - 1]
            modules += [nn.Linear(input_size, units), nn.ReLU()]
            if self.batch_normalization:
                modules.append(nn.BatchNorm1d(units))
        if len(dims) == 1:
            modules.append(
                nn.Linear(context_length, dims[-1] * prediction_length)
            )
        else:
            modules.append(nn.Linear(dims[-2], dims[-1] * prediction_length))
        modules.append(
            LambdaLayer(
                lambda o: torch.reshape(o, (-1, prediction_length, dims[-1]))
            )
        )
        self.mlp = nn.Sequential(*modules)

        self.distr_args_proj = self.distr_output.get_args_proj(dims[-1])

        self.scaler = MeanScaler() if mean_scaling else NOPScaler()

    def get_distr(self, past_target: torch.Tensor) -> Distribution:
        # (batch_size, seq_len, target_dim) and (batch_size, seq_len, target_dim)
        scaled_target, target_scale = self.scaler(
            past_target,
            torch.ones_like(
                past_target
            ),  # TODO: pass the actual observed here
        )

        mlp_outputs = self.mlp(scaled_target)
        distr_args = self.distr_args_proj(mlp_outputs)
        return self.distr_output.distribution(
            distr_args, scale=target_scale.unsqueeze(1)
        )


class SimpleFeedForwardTrainingNetwork(SimpleFeedForwardNetworkBase):
    def forward(
        self, past_target: torch.Tensor, future_target: torch.Tensor
    ) -> torch.Tensor:
        distr = self.get_distr(past_target)

        # (batch_size, prediction_length, target_dim)
        loss = -distr.log_prob(future_target)

        return loss.mean()


class SimpleFeedForwardPredictionNetwork(SimpleFeedForwardNetworkBase):
    @validated()
    def __init__(
        self, num_parallel_samples: int = 100, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.num_parallel_samples = num_parallel_samples

    def forward(self, past_target: torch.Tensor) -> torch.Tensor:
        distr = self.get_distr(past_target)

        # (num_samples, batch_size, prediction_length)
        samples = distr.sample((self.num_parallel_samples,))

        return samples.permute(1, 0, 2)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional, Tuple, List, Union
from itertools import product
from pathlib import Path

from torch.utils.data import Dataset as TorchDataset
from joblib import Parallel, delayed
from tqdm import tqdm
from torch import Tensor
import torch as pt
import pandas as pd
import numpy as np

from tslib.dataset import (
    TimeSeries,
    TimeSeriesCorpus,
    WindowsDataset,
    MetaDataset,
)
from data import (
    SeasonalReader,
    FourierSignalReader,
    FusedSeasonalSignalReader,
    RandomSeasonalSignalReader,
    TrendFourierSignalReader,
    GluonTSJsonReader,
)


def _tensor_collator(batch: List[Tensor]):
    elem = batch[0]
    lengths = [t.size(0) for t in batch]
    max_length = max(lengths)
    pad_size = [max_length - l for l in lengths]

    storage = elem.storage()._new_shared(
        len(batch) * max_length * elem[0].numel()
    )
    out = elem.new(storage)
    if max(pad_size) == 0:
        tensor = pt.stack(batch, dim=0, out=out)
        lengths = None
    else:
        # shorter sequences are padded at left
        tensor = pt.stack(
            [
                pt.cat(
                    [sample.new_zeros(pad, *sample.shape[1:]), sample], dim=0
                )
                for pad, sample in zip(pad_size, batch)
            ],
            dim=0,
            out=out,
        )
        lengths = tensor.new_tensor(lengths, dtype=pt.long)
    return tensor, lengths


def _variable_length_collator(batch: List[Tuple]):
    data, feats, nan_mask = zip(*batch)
    data, length = _tensor_collator(data)
    if feats[0] is not None:
        feats, _ = _tensor_collator(feats)
    else:
        feats = None
    if nan_mask[0] is not None:
        nan_mask, _ = _tensor_collator(nan_mask)
    else:
        nan_mask = None
    return data, feats, nan_mask, length


def _pair_variable_length_collator(batch: List[Tuple]):
    (
        src_data,
        tgt_data,
        src_feats,
        tgt_feats,
        src_nan_mask,
        tgt_nan_mask,
    ) = zip(*batch)
    if src_data[0] is None:
        src_input = (None, None, None, None)
    else:
        src_input = _variable_length_collator(
            list(zip(src_data, src_feats, src_nan_mask))
        )
    if tgt_data[0] is None:
        tgt_input = (None, None, None, None)
    else:
        tgt_input = _variable_length_collator(
            list(zip(tgt_data, tgt_feats, tgt_nan_mask))
        )
    return sum(zip(src_input, tgt_input), ())


class DomAdaptDataset(TorchDataset):
    def __init__(
        self,
        source_dataset: Optional[WindowsDataset],
        target_dataset: Optional[WindowsDataset],
    ) -> None:
        super(DomAdaptDataset, self).__init__()
        if (source_dataset is None) and (target_dataset is None):
            raise ValueError("Source and target cannot be both None.")
        self.source_dataset = source_dataset
        self.target_dataset = target_dataset
        self._source_size = (
            1 if self.source_dataset is None else len(self.source_dataset)
        )
        self._target_size = (
            1 if self.target_dataset is None else len(self.target_dataset)
        )

    def __len__(self) -> int:
        return self._target_size

    def __getitem__(self, index: int) -> Tuple:
        src_index = np.random.choice(self._source_size)
        tgt_index = index
        src = (
            None
            if self.source_dataset is None
            else self.source_dataset[src_index]
        )
        tgt = (
            None
            if self.target_dataset is None
            else self.target_dataset[tgt_index]
        )
        if src is None:
            src = tuple([None] * len(tgt))
        if tgt is None:
            tgt = tuple([None] * len(src))
        return sum(zip(src, tgt), ())

    @classmethod
    def from_domains(
        cls,
        source_dataset: Optional[MetaDataset],
        target_dataset: Optional[MetaDataset],
    ) -> MetaDataset:
        def _get_subset(
            dataset: Optional[MetaDataset], name: str
        ) -> Optional[WindowsDataset]:
            if dataset is None:
                return None
            return getattr(dataset, name)

        kwargs = {}
        kwargs["train_set"] = cls(
            _get_subset(source_dataset, "train_set"),
            _get_subset(target_dataset, "train_set"),
        )
        kwargs["test_set"] = cls(
            _get_subset(source_dataset, "test_set"),
            _get_subset(target_dataset, "test_set"),
        )
        try:
            kwargs["valid_set"] = cls(
                _get_subset(source_dataset, "valid_set"),
                _get_subset(target_dataset, "valid_set"),
            )
        except ValueError:
            kwargs["valid_set"] = None

        if (source_dataset is not None) and (
            source_dataset.collate_fn is _variable_length_collator
        ):
            kwargs.setdefault("collate_fn", _pair_variable_length_collator)
        if (target_dataset is not None) and (
            target_dataset.collate_fn is _variable_length_collator
        ):
            kwargs.setdefault("collate_fn", _pair_variable_length_collator)

        return MetaDataset(**kwargs)


class SeasonalDataset(WindowsDataset, SeasonalReader):
    def __getitem__(self, index: int) -> Tuple[Tensor]:
        ts = super(SeasonalDataset, self).__getitem__(index)
        data = pt.tensor(ts.target, dtype=pt.float)
        return (data,)

    @property
    def d_data(self) -> int:
        return 1

    @property
    def d_feats(self) -> int:
        return 0

    @classmethod
    def full_windows(cls, corpus: TimeSeriesCorpus):
        windows = [(i, 0, len(ts)) for i, ts in enumerate(corpus)]
        return cls(corpus, windows)

    @classmethod
    def create_dataset(
        cls,
        n_instances_train: int,
        n_instances_test: int,
        n_samples,
        *freqs,
        **kwargs,
    ):
        train_corpus = cls.load_data(
            n_instances_train, n_samples, *freqs, **kwargs
        )
        valid_corpus = cls.load_data(
            n_instances_test, n_samples, *freqs, **kwargs
        )
        test_corpus = cls.load_data(
            n_instances_test, n_samples, *freqs, **kwargs
        )
        dataset = MetaDataset(
            train_set=cls.full_windows(train_corpus),
            valid_set=cls.full_windows(valid_corpus),
            test_set=cls.full_windows(test_corpus),
        )
        return dataset


class FourierDataset(SeasonalDataset, FourierSignalReader):
    pass


class ComplexDataset(SeasonalDataset, FusedSeasonalSignalReader):
    pass


class RandomDataset(SeasonalDataset, RandomSeasonalSignalReader):
    pass


class TrendFourierDataset(SeasonalDataset, TrendFourierSignalReader):
    pass


class BenchmarkDataset(WindowsDataset, GluonTSJsonReader):
    def __getitem__(
        self, index: int
    ) -> Tuple[Tensor, Optional[Tensor], Optional[Tensor]]:
        window = super(BenchmarkDataset, self).__getitem__(index)
        target = window.target
        nan_mask = np.isnan(target).any(axis=1)
        nan_mask = pt.tensor(nan_mask, dtype=pt.bool)
        target = np.nan_to_num(target)
        target = pt.tensor(target, dtype=pt.float)
        feat_names = sorted(window.revealed_numerical_features.keys())
        if len(feat_names) > 0:
            feats = pt.tensor(
                np.concatenate(
                    [
                        window.revealed_numerical_features[k]
                        for k in feat_names
                    ],
                    axis=1,
                ),
                dtype=pt.float,
            )
        else:
            feats = None

        return target, feats, nan_mask

    @property
    def d_data(self) -> int:
        return 1

    @property
    def d_feats(self) -> int:
        instance = self.corpus.instances[0]
        keys = instance.revealed_numerical_features.keys()
        return sum(instance._features[k].shape[-1] for k in keys)

    @classmethod
    def create_dataset(
        cls,
        dataset: str,
        freq: str,
        n_instances_train: int,
        n_instances_eval: int,
        eval_size: int,
        window_size: int,
        window_size_plus: int = 0,
        window_size_minus: int = 0,
        window_shift: int = 1,
        train_size: Optional[int] = None,
        validation: bool = True,
        seed: int = 42,
    ) -> MetaDataset:
        corpus = cls.load_data(dataset, freq)
        train_corpus, test_corpus = corpus.split_from_end(eval_size, 0)
        if validation:
            train_corpus, valid_corpus = train_corpus.split_from_end(
                eval_size, 0, raise_error=False
            )
        else:
            valid_corpus = None
        if train_size is not None:
            _, train_corpus = train_corpus.split_from_end(
                train_size, 0, raise_error=False
            )

        kwargs = {}
        kwargs["test_set"] = cls.random_windows(
            corpus=test_corpus,
            n_windows=n_instances_eval,
            window_size=window_size,
            window_size_plus=window_size_plus,
            window_size_minus=window_size_minus,
            window_shift=window_shift,
            seed=seed,
        )
        if valid_corpus is not None:
            kwargs["valid_set"] = cls.random_windows(
                corpus=valid_corpus,
                n_windows=n_instances_eval,
                window_size=window_size,
                window_size_plus=window_size_plus,
                window_size_minus=window_size_minus,
                window_shift=window_shift,
                seed=seed,
            )
        else:
            kwargs["valid_set"] = None
        kwargs["train_set"] = cls.random_windows(
            corpus=train_corpus,
            n_windows=n_instances_train,
            window_size=window_size,
            window_size_plus=window_size_plus,
            window_size_minus=window_size_minus,
            window_shift=window_shift,
            seed=seed,
        )
        if window_size_plus > 0 or window_size_minus > 0:
            kwargs["collate_fn"] = _variable_length_collator
        return MetaDataset(**kwargs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import re
from datetime import datetime


def make_metrics(metrics_names):
    avg_epoch_loss_metric = {
        "Name": "training_loss",
        "Regex": r"'epoch_loss'=(\S+)",
    }
    final_loss_metric = {"Name": "final_loss", "Regex": r"Final loss: (\S+)"}
    other_metrics = [
        {
            "Name": metric,
            "Regex": rf"gluonts\[metric-{re.escape(metric)}\]: (\S+)",
        }
        for metric in metrics_names
    ]

    return [avg_epoch_loss_metric, final_loss_metric] + other_metrics


def make_job_name(base_job_name):
    now = datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%S-%f")
    return f"{base_job_name}-{now}"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import torch
import numpy as np


def tensor_to_np(tensor: torch.Tensor) -> np.ndarray:
    return tensor.cpu().detach().numpy()


def get_mask(pad_size: torch.Size, lengths: torch.Tensor) -> torch.Tensor:
    device = lengths.device
    return torch.arange(pad_size, device=device)[None, :] < lengths[
        :, None
    ].to(device)
